﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Linq;
using System.Text;


namespace ITIN.CSharp.BaseObjects
{
    /// <summary>
    /// Интерфейс уведомлений об изменениях стакана
    /// </summary>
    public interface INotifyDOMUpdated
    {
        /// <summary>
        /// Уведомить об изменении всего стакана
        /// </summary>
        void NotifyDOMUpdated();

        /// <summary>
        /// Уведомить об изменение покупок
        /// </summary>
        void NotifyBidsUpdated();

        /// <summary>
        /// Уведомить об изменение продаж
        /// </summary>
        void NotifyAsksUpdated();
    }

    /// <summary>
    /// Интерфейс ядра торговой системы
    /// </summary>
    public interface IEngine
    {
        /// <summary>
        /// Получить биржу по имени
        /// </summary>
        /// <param name="exchangeName">Имя биржи</param>
        /// <returns>Экземпляр биржи</returns>
        Exchange GetExchange(string exchangeName);
        
        /// <summary>
        /// Получить торговый счет
        /// </summary>
        /// <param name="exchange">Наименование счёта</param>
        /// <returns>Экземпляр счёта</returns>
        List<Account> GetAccounts(Exchange exchange);
        
        /// <summary>
        /// Получить торговый инструмент
        /// </summary>
        /// <param name="ticker">Наименование инструмента</param>
        /// <returns>"Экемпляр инструмента</returns>
        Ticker GetTicker(string symbol);
        
        /// <summary>
        /// Получить экземпляр стратегии
        /// </summary>
        /// <param name="name">Имя стратегии</param>
        /// <returns>Экземпляр стратегии</returns>
        IStrategy GetStrategy(string name);

        /// <summary>
        /// Создать стратегию
        /// </summary>
        /// <param name="strategyType">Тип Класса создаваемой стратегии</param>
        /// <param name="name">Имя стратегии</param>
        /// <returns>"Экземпляр стратегии</returns>
        IStrategy CreateOrGetStrategy(Type strategyType, string name);
        

        /// <summary>
        /// Отправить заявку
        /// </summary>
        /// <param name="order">Экземпляр ордера, хранит параметры для отправки на биржу</param>
        void SendOrder(TradeOrder order);
        
        /// <summary>
        /// Отмена(снятие) заявки
        /// </summary>
        /// <param name="order">Экземпляр ордера, хранит параметры для отправки на биржу</param>
        void CancelOrder(TradeOrder order);
        
        /// <summary>
        /// Перемещение заявки(ордера)
        /// </summary>
        /// <param name="order">Экземпляр ордера, хранит параметры для отправки на биржу</param>
        /// <param name="newPrice">Новая цена заявки</param>
        void MoveOrder(TradeOrder order, double newPrice);
    }

    /// <summary>
    /// Интерфейс стратегии
    /// </summary>
    public interface IStrategy
    {
        /// <summary>
        /// Возвращает позицию стратегии по указанному инструменту
        /// </summary>
        /// <param name="ticker"></param>
        /// <returns>Текущая позиция</returns>
        double GetPosition(Ticker ticker);

        /// <summary>
        /// Строка, перечисляющая все позиции
        /// </summary>
        string Positions { get; } 

        /// <summary>
        /// Поместить в очередь стратегии данные(задание) на обработку в потоке стратегии
        /// </summary>
        /// <param name="data">Данные для обработки</param>
        void SendStrategyData(Array data);

        /// <summary>
        /// Поместить в очередь стратегии SGH задание на исполнение в потоке стратегии
        /// </summary>
        /// <param name="action">Действие, котороенеобходимо выполнить в потоке стратегии</param>
        void SendStrategyAction(Action action);

        /// <summary>
        /// Останов стратегии
        /// </summary>
        void Start();
        
        /// <summary>
        /// Запуск Стратегии
        /// </summary>
        /// <returns>Задача запущенной стратегии</returns>
        System.Threading.Tasks.Task Stop();

        /// <summary>
        /// Свойство возвращает запущена ли стратегия
        /// </summary>
        bool IsStarted { get; }

        /// <summary>
        /// Имя стратегии
        /// </summary>
        string Name { get; }
    }

}
