﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;
using System.Threading;

using ITIN.CSharp.Types;

namespace ITIN.CSharp.BaseObjects
{
    /// <summary>
    /// Класс представляет собой ценовой уровень с 3 параметрами 
    /// - Цена
    /// - Количество
    /// - Свой объем в заявках по данной цене
    /// </summary>
    public class Level
    {
        #region Methods

        /// <summary>
        /// Конструктор
        /// </summary>
        public Level()
        {
            this.Price = 0.0;
            this.Amount = 0.0;
            this.OwnAmount = 0.0;
        }

        /// <summary>
        /// Конструктор с параметрами
        /// </summary>
        /// <param name="price">Цена</param>
        /// <param name="alienAmount">Количество</param>
        /// <param name="ownAmount">Объем своих заявок</param>
        public Level(double price, double alienAmount, double ownAmount = 0.0)
        {
            this.Price = price;
            this.Amount = alienAmount;
            this.OwnAmount = ownAmount;
        }

        #endregion Methods

        #region Properties

        /// <summary>
        /// Цена
        /// </summary>
        public double Price { get; set; }
        /// <summary>
        /// Количество
        /// </summary>
        public double Amount { get; set; }
        /// <summary>
        /// Своё количество
        /// </summary>
        public double OwnAmount { get; set; }

        /// <summary>
        /// Являеться ли уровень продажей
        /// </summary>
        public bool IsAsk { get { return Amount < 0.0; } }
        
        /// <summary>
        /// Являеться ли уровень покупкой
        /// </summary>
        public bool IsBid { get { return Amount > 0.0; } }
        
        /// <summary>
        /// Являеться ли уровень нейтральным (объем равен нулю)
        /// </summary>
        public bool IsFlat { get { return Amount == 0.0; } }
        
        /// <summary>
        /// Стоит ли на данном уровне наша заявка
        /// </summary>
        public bool IsOwn { get { return OwnAmount != 0.0; } }

        #endregion Properties

        #region Interface Implementation

        /// <summary>
        /// Выводит строковое представление ценового уровня
        /// </summary>
        /// <returns>Строка ценового уровня</returns>
        public override string ToString()
        {
            return this.Price.ToString() + " | " + this.Amount;
        }

        #endregion Interface Implementation
    }
    
    /// <summary>
    /// Класс, снимок очереди покупок или продаж
    /// </summary>
    public class SnapshotDOM
    {
        #region Fields

        /// <summary>
        /// Список заявок
        /// </summary>
        List<Level> _snapshot;

        /// <summary>
        /// эффективный класс для блокировок read/write
        /// </summary> 
        private ReaderWriterLockSlim _locker;
        
        #endregion fields

        #region Methods

        public SnapshotDOM()
        {
            this._snapshot = new List<Level>();
            this._locker = new ReaderWriterLockSlim(LockRecursionPolicy.SupportsRecursion);
        }

        public Level this[int index]
        {
            get
            {
                if (index < 0) return null;
                Level tmpLevel = null;
                this._locker.EnterReadLock();
                try
                {
                    if ((index >= 0) && (index < this._snapshot.Count))
                    {
                        tmpLevel = this._snapshot[index];
                    }
                }
                finally
                {
                    this._locker.ExitReadLock();
                }
                return tmpLevel;
            }
        }

        public bool AddUpdate(int index, double price, double amount)
        {
            if (index < 0) return false;
            
            bool isUpdate = false;
            this._locker.EnterUpgradeableReadLock();
            try
            {
                if (index < _snapshot.Count)
                {
                    if ((this._snapshot[index].Price != price) ||
                       (this._snapshot[index].Amount != amount))
                    {
                        isUpdate = true;

                        this._locker.EnterWriteLock();
                        try
                        {
                            this._snapshot[index].Price = price;
                            this._snapshot[index].Amount = amount;
                        }
                        finally
                        {
                            this._locker.ExitWriteLock();
                        }
                    }
                }
                else
                { // index >= this._snaphot.Count
                    int newLevelCount = (index + 1) - this._snapshot.Count;
                    for (int i = 0; i < newLevelCount; i++)
                    {
                        Level newLevel = null;
                        if (this._snapshot.Count + i == index)
                        {
                            newLevel = new Level(price, amount);
                        }
                        else
                        {
                            newLevel = new Level(0.0, 0.0);
                        }
                        
                        this._locker.EnterWriteLock();                        
                        try
                        {
                            this._snapshot.Add(newLevel);
                        }
                        finally
                        {
                            this._locker.ExitWriteLock();                            
                        }
                    }
                    isUpdate = true;
                }
            }
            finally
            {
                this._locker.ExitUpgradeableReadLock();
            }
            return isUpdate;
        }

        public void ClearSnapshotTail(int startIndexToClear)
        {
            this._locker.EnterUpgradeableReadLock();
            try
            {
                for (int i = startIndexToClear; i < this._snapshot.Count; i++)
                {
                    if (this._locker.IsReadLockHeld == false)
                    {
                        this._locker.EnterWriteLock();
                    }
                    this._snapshot[i].Amount = 0.0;
                    this._snapshot[i].Price = 0.0;
                    // не трогаем количество в своих заявках
                    //this._snaphot[i].OwnAmount
                }
            }
            finally
            {
                if (this._locker.IsReadLockHeld == true)
                {
                    this._locker.ExitWriteLock();
                }
                this._locker.EnterUpgradeableReadLock();
            }
        }

        #endregion Methods

        #region Properties

        /// <summary>
        /// Количество заявок в очереди
        /// </summary>
        public int Count { get { return this._snapshot.Count; } }

        /// <summary>
        /// Лучшая цена
        /// </summary>
        public double BestPrice
        {
            get
            {
                // вернем ноль для упрощения
                if (Count == 0) return 0.0; // по хорошему нужно возвращать минмиальные и максимальные значения

                return this._snapshot[0].Price;
            }
        }


        #endregion Properties

        #region Interface Implementation

        /*public override string ToString()
        {
            return 
        }*/

        #endregion Interface Implementation

    }
    
    /// <summary>
    /// Стакан заявок (очередь покупок и продаж)
    /// </summary>
    public class DepthOfMarket
    {
        #region Fields

        /// <summary>
        /// Покупки
        /// </summary>
        private SnapshotDOM _buys;
        
        /// <summary>
        /// Продажи
        /// </summary>
        private SnapshotDOM _sells;

        /// <summary>
        /// визуальное представление стакана покупок
        /// </summary>
        private ObservableCollection<Level> _externalVisualBids;
        
        /// <summary>
        /// визуальное представление стакана продаж
        /// </summary>
        private ObservableCollection<Level> _externalVisualAsks;
        
        #endregion fields

        #region Methods

        public DepthOfMarket(Ticker ticker)
        {
            this._buys = new SnapshotDOM();            
            this._sells = new SnapshotDOM();

            this.Ticker = ticker;
        }

        #endregion Methods

        /// <summary>
        /// Обновление очереди покупок
        /// </summary>
        /// <param name="index">Индекс в очереди</param>
        /// <param name="price">Цена</param>
        /// <param name="amount">Количество</param>
        public void UpdateBidQuote(int index, double price, double amount)
        {
            bool isUpdated = this._buys.AddUpdate(index, price, amount);
        }

        /// <summary>
        /// Обновление очереди продаж
        /// </summary>
        /// <param name="index">Индекс в очереди</param>
        /// <param name="price">Цена</param>
        /// <param name="amount">Количество</param>
        public void UpdateAskQuote(int index, double price, double amount)
        {
            bool isUpdated = this._sells.AddUpdate(index, price, amount);
        }

        /// <summary>
        /// Связать внешние ссылки на текущие визуальные представления стакана
        /// </summary>
        /// <param name="externalVisualBids"></param>
        /// <param name="externalVisualAsks"></param>
        public void AssignWithVisualDOM(ObservableCollection<Level> externalVisualBids, ObservableCollection<Level> externalVisualAsks)
        {
            this._externalVisualBids = externalVisualBids;
            this._externalVisualAsks = externalVisualAsks;
        }

        #region Properties

        /// <summary>
        /// Инструмент
        /// </summary>
        public Ticker Ticker { get; private set; }

        /// <summary>
        /// Покупки
        /// </summary>
        public SnapshotDOM Bids { get { return this._buys; } }

        /// <summary>
        /// Продажи
        /// </summary>
        public SnapshotDOM Asks { get { return this._sells; } }

        #endregion Properties

        #region Interface Implementation

        /// <summary>
        /// Возвращает строкове представление стакана
        /// </summary>
        /// <returns>Строка, предствляющая стакан</returns>
        public override string ToString()
        {
            StringBuilder res = new StringBuilder();
            if (this._buys.Count > 0)
            {
                res.Append(this._buys[0].Price.ToString());
            }
            else
            {
                res.Append("0.0");
            }
            res.Append(" / ");

            if (this._sells.Count > 0)
            {
                res.Append(this._sells[0].Price.ToString());
            }
            else
            {
                res.Append("0.0");
            }
            return res.ToString();
        }

        #endregion Interface Implementation

    }
}
