﻿using System;
using System.Collections.Generic;
using System.Collections.Concurrent;
using System.Collections.Specialized;
using System.Globalization;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading;

//using System.Winwods.Threading;

namespace ITIN.CSharp.BaseObjects
{
    /// <summary>
    /// Класс реализует подписку на события по шаблону Observer
    /// </summary>
    /// <typeparam name="T"></typeparam>
    public class SafeObservableEvents<T>
    {
        #region Fields

        /// <summary>
        /// Блокировка
        /// </summary>
        private ReaderWriterLockSlim _lock;
        
        /// <summary>
        /// Список подписчиков
        /// </summary>
        private  List<SafeObserverEvent<T>> _observers;

        #endregion Fields

        #region Methods

        /// <summary>
        /// Конструктор
        /// </summary>
        public SafeObservableEvents()
        {
            this._lock = new ReaderWriterLockSlim(LockRecursionPolicy.SupportsRecursion);
            this._observers = new List<SafeObserverEvent<T>>();
        }

        /// <summary>
        /// Уведомить о событии подписчиков
        /// </summary>
        /// <param name="obj">Параметр события</param>
        public void Notify(T obj)
        {
            _lock.EnterReadLock();
            try
            {
                foreach (var observer in _observers)
                {
                    observer.DoAction(obj);
                }
            }
            finally
            {
                this._lock.ExitReadLock();
            }
        }

        /// <summary>
        /// Подписаться на рассылку уведомлений о событии
        /// </summary>
        /// <param name="action">Действие на подписку</param>
        /// <returns>Индерфейс для отписки от события</returns>
        public IDisposable Subscribe(Action<T> action)
        {
            SafeObserverEvent<T> observer = null;
            this._lock.EnterUpgradeableReadLock();
            try
            {
                observer = new SafeObserverEvent<T>(action);
                if (this._observers.Contains(observer) == false)
                {
                    this._lock.EnterUpgradeableReadLock();
                    try
                    {
                        this._observers.Add(observer);
                    }
                    finally
                    {
                        this._lock.ExitUpgradeableReadLock();
                    }
                }
            }
            finally
            {
                this._lock.ExitUpgradeableReadLock();
            }

            return new Unsubscriber(this, observer);
        }

        #endregion Methods

        /// <summary>
        /// Событие с уведомлением и отпиской
        /// </summary>
        /// <typeparam name="T"></typeparam>
        private class SafeObserverEvent<T>
        {
            #region Fields
            
            /// <summary>
            /// Блокировка
            /// </summary>
            internal ReaderWriterLockSlim _lock;

            /// <summary>
            /// Действие
            /// </summary>
            internal Action<T> _eventAction;

            #endregion Methods

            #region Methods

            /// <summary>
            /// Конструктор
            /// </summary>
            /// <param name="eventAction">Действие по событию</param>
            internal SafeObserverEvent(Action<T> eventAction)
            {
                this._lock = new ReaderWriterLockSlim(LockRecursionPolicy.SupportsRecursion);
                this._eventAction = eventAction;
            }

            /// <summary>
            /// Выполнить дейсивие подписки
            /// </summary>
            /// <param name="val"></param>
            internal void DoAction(T val)
            {
                this._lock.EnterReadLock();
                try
                { 
                    if (_eventAction != null)
                        this._eventAction(val);
                }
                finally
                {
                    this._lock.ExitReadLock();
                }
            }

            #endregion Methods
        }

        /// <summary>
        /// Класс отписчик от события
        /// </summary>
        private class Unsubscriber : IDisposable
        {
            #region Fields

            /// <summary>
            /// События
            /// </summary>
            private SafeObservableEvents<T> _observers;
            
            /// <summary>
            /// Уведомитель о событии
            /// </summary>
            private SafeObserverEvent<T> _observer;

            #endregion Methods

            #region Methods

            /// <summary>
            /// Конструктор класса отписчика
            /// </summary>
            /// <param name="observers">Класс-список событий</param>
            /// <param name="observer">Класс уведомитель о собитии</param>
            public Unsubscriber(SafeObservableEvents<T> observers, SafeObserverEvent<T> observer)
            {
                this._observers = observers;
                this._observer = observer;
            }

            /// <summary>
            /// Отписаться от события
            /// </summary>
            public void Dispose()
            {
                this._observers._lock.EnterUpgradeableReadLock();
                try
                {
                    this._observer._lock.EnterWriteLock();
                    try
                    {
                        if (_observer != null)
                        {
                            // обработчик освободился, пока обнулим его чтоб больше и не исполнялся
                            _observer._eventAction = null;
                        }
                    }
                    finally
                    {
                        this._observer._lock.ExitWriteLock();
                    }

                    if (_observers._observers.Contains(_observer) == true)
                    {
                        this._observers._lock.EnterWriteLock();
                        try
                        {
                            _observers._observers.Remove(_observer);
                        }
                        finally
                        {
                            this._observers._lock.ExitWriteLock();
                        }
                    }
                }
                finally
                {
                    this._observers._lock.EnterUpgradeableReadLock();
                }
            }
            
            #endregion Methods
        }
    }
}
