﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Linq;
using System.Text;

namespace ITIN.CSharp.BaseObjects
{
    /// <summary>
    /// Потокобезопасный класс словаря с возможностью отображения изменений на контролах коллекций WPF
    /// </summary>
    /// <typeparam name="TKey">Тип ключа</typeparam>
    /// <typeparam name="TValue">Тип значения</typeparam>
    public partial class SafeObservableDictionary<TKey, TValue> : ConcurrentDictionary<TKey, TValue>, INotifyCollectionChanged
    {
        /// <summary>
        /// Внешнний преобработчик уведомления, необходим например для потокобезопасного обновления контрола
        /// </summary>
        private Action<Action> _externalCollectionChangeHandler = null; 


        public SafeObservableDictionary() : base() { ExternalCollectionChangeHandler = null; }
        public SafeObservableDictionary(IEqualityComparer<TKey> comparer) : base(comparer) { }
        public SafeObservableDictionary(IDictionary<TKey, TValue> dictionary) : base(dictionary) { }
        public SafeObservableDictionary(IDictionary<TKey, TValue> dictionary, IEqualityComparer<TKey> comparer) : base(dictionary, comparer) { }

        /// <summary>
        /// Событие уведомления об изменении
        /// </summary>
        public event NotifyCollectionChangedEventHandler CollectionChanged;

        /// <summary>
        /// Индексатор
        /// </summary>
        /// <param name="key">Ключ</param>
        /// <returns>Значение по ключу</returns>
        public new TValue this[TKey key]
        {
            get
            {
                return base[key];
            }
            set
            {
                OnCollectionChanged(new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Replace, key, 0));
                base[key] = value;
            }
        }

        /// <summary>
        /// Пытаеться добавить отсутсвтующее значение в коллекцию по ключу
        /// </summary>
        /// <param name="key">Ключ</param>
        /// <param name="value">Значение</param>
        /// <returns>Возвращается значение о присутствии ключа</returns>
        public new bool TryAdd(TKey key, TValue value)
        {
            bool isNewElement = base.TryAdd(key, value);

            if (isNewElement == true) 
                OnCollectionChanged(new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Add, key, 0));
            
            return isNewElement;
        }

        /// <summary>
        /// Добавляет или изменяет значение
        /// </summary>
        /// <param name="key">Ключ</param>
        /// <param name="value">Значение</param>
        /// <returns>Элемент</returns>
        public new TValue GetOrAdd(TKey key, TValue value)
        {
            bool isNewElement = false;
            TValue resultingValue = value;
            base.AddOrUpdate(key,
                (keyAdd) =>
                {
                    isNewElement = true;
                    return value; //resultingValue = value;
                },
                (keyUpdate, valPrev) =>
                {   
                    resultingValue = valPrev;
                    return valPrev;
                });

            if (isNewElement == true)
                OnCollectionChanged(new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Add, key, 0));
            
            return resultingValue;
        }

        /// <summary>
        /// Добавляет или изменяет значение
        /// </summary>
        /// <param name="key">Ключ</param>
        /// <param name="valueFactory">Функция создания элемента</param>
        /// <returns>Элемент</returns>
        public new TValue GetOrAdd(TKey key, Func<TKey, TValue> valueFactory)
        {
            bool isNewElement = false;
            TValue resultingValue = default(TValue);
            base.AddOrUpdate(key,
                (keyAdd) =>
                {
                    isNewElement = true;
                    resultingValue = valueFactory(keyAdd);
                    return resultingValue;
                },
                (keyUpdate, valPrev) =>
                {
                    return resultingValue = valPrev;
                });

            if (isNewElement == true)
                OnCollectionChanged(new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Add, key, 0));

            return resultingValue;
        }

        /// <summary>
        /// Добавить или изменить значение по ключу
        /// </summary>
        /// <param name="key">Ключ</param>
        /// <param name="value">Значение</param>
        /// <param name="updateValueFactory">Функция возвращает значение, которым необходимо заменияеться существующее</param>
        /// <returns>Новый элемент по ключу</returns>
        public new TValue AddOrUpdate(TKey key, TValue value, Func<TKey, TValue, TValue> updateValueFactory)
        {
            NotifyCollectionChangedEventArgs notifyChange = null;
            TValue resultingValue = value;

            base.AddOrUpdate(key,
                (keyAdd) =>
                {
                    resultingValue = value;
                    notifyChange = new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Add, keyAdd, 0);
                    return resultingValue;
                },
                (keyUpdate, valPrev) =>
                {
                    resultingValue = updateValueFactory(keyUpdate, valPrev);
                    notifyChange = new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Replace, keyUpdate, 0);
                    return resultingValue = valPrev;
                });
            if (notifyChange != null) OnCollectionChanged(notifyChange);
            
            return resultingValue;
        }

        /// <summary>
        /// Добавить или изменить значение по ключу
        /// </summary>
        /// <param name="key">Ключ</param>
        /// <param name="addValueFactory">Функция возвращает новое значение</param>
        /// <param name="updateValueFactory">Функция возвращает значение, которым необходимо заменияеться существующее</param>
        /// <returns>Новый элемент по ключу</returns>
        public new TValue AddOrUpdate(TKey key, Func<TKey, TValue> addValueFactory, Func<TKey, TValue, TValue> updateValueFactory)
        {
            NotifyCollectionChangedEventArgs notifyChange = null;
            TValue resultingValue = default(TValue);

            base.AddOrUpdate(key,
                (keyAdd) =>
                {
                    resultingValue = addValueFactory(keyAdd);
                    notifyChange = new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Add, keyAdd, 0);
                    return resultingValue;
                },
                (keyUpdate, valPrev) =>
                {
                    resultingValue = updateValueFactory(keyUpdate, valPrev);
                    notifyChange = new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Replace, keyUpdate, 0);
                    return resultingValue = valPrev;
                });
            if (notifyChange != null) OnCollectionChanged(notifyChange);

            return resultingValue;
        }

        /// <summary>
        /// Попытка заменить значение по ключу
        /// </summary>
        /// <param name="key">Ключ</param>
        /// <param name="newValue">Новое значение</param>
        /// <param name="comparsionValue">Значение, которое сравнивается со значением элемента с ключом key.</param>
        /// <returns>Значение true, если значение с ключом key оказалось равно значению параметра comparisonValue и было заменено значением newValue; в противном случае — значение false</returns>
        public new bool TryUpdate(TKey key, TValue newValue, TValue comparsionValue)
        {
            bool result = base.TryUpdate(key, newValue, comparsionValue);
            if (result == true)
                OnCollectionChanged(new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Replace, key, 0));

            return result;
        }

        /// <summary>
        /// Пытается удалить и вернуть значение с указанным ключом из коллекции.
        /// </summary>
        /// <param name="key">Ключ удаляемого и возвращаемого элемента.</param>
        /// <param name="deletedValue">Удаленное значение</param>
        /// <returns>Значение true, если объект был успешно удален; в противном случае — значение false.</returns>
        public new bool TryRemove(TKey key, out TValue deletedValue)
        {
            bool result = base.TryRemove(key, out deletedValue);
            if (result == true)
            {
                //if (base.Count == 0.0)
                {
                    OnCollectionChanged(new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Reset));
                }
                //else
                {
                    //OnCollectionChanged(new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Remove, key, 0));
                }
            }
            
            return result;
        }

        /// <summary>
        /// Очистка коллекции
        /// </summary>
        public new void Clear()
        {
            base.Clear();
            OnCollectionChanged(new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Reset));
        }

        /// <summary>
        /// Внешний преобработчик обновления
        /// </summary>
        public Action<Action> ExternalCollectionChangeHandler
        { 
            get
            {
                return this._externalCollectionChangeHandler;
            }
            set
            {
                this._externalCollectionChangeHandler = value;
            }
        }

        /// <summary>
        /// Уведомление об изменении
        /// </summary>
        /// <param name="e">Параметры изменений</param>
        protected virtual void OnCollectionChanged(NotifyCollectionChangedEventArgs e)
        {
            NotifyCollectionChangedEventArgs eLocal = new NotifyCollectionChangedEventArgs(e.Action, e.NewItems, e.NewStartingIndex);
 
            Action collectionChanged = () => 
                {
                    NotifyCollectionChangedEventArgs eLocalInActionBlock = eLocal;
                    if (CollectionChanged != null)
                    {
                        CollectionChanged(this, eLocalInActionBlock);
                    }
                };

            if (_externalCollectionChangeHandler != null)
            {
                this._externalCollectionChangeHandler(collectionChanged);
            }
            else
            {
                collectionChanged();
            }
        }

    }
}
