﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Globalization;
using System.Linq;
using System.Text;

using ITIN.CSharp.Types;

namespace ITIN.CSharp.BaseObjects
{
    /// <summary>
    /// Класс, представляющий экземпляр торгового инструмента
    /// </summary>
    public class Ticker
    {
        #region Fields

        private string   _symbol;            // код бумаги
        private string   _shortName;         // короткое название
        private string   _longName;          // полное наименование инструмента
        private double   _step;              // шаг цены
        private double   _stepPrice;         // Стоимость шага цены
        private int      _decimals;          // число знаков после запятой
        private string   _secExtId;          // идентификатор биржи
        private DateTime _expiryDate;        // дата экспирации
        private double   _daysBeforeExpiry;  // дней до экспирации
        private Exchange _exchange;          // биржа

        private QuoteCommon _common;         // Общая информация по инструменту

        private DepthOfMarket _dom;          // "Стакан" рыночных спросов и предложений

        private ObservableCollection<Deal> _deals; // Список сделок

        private SafeObservableEvents<Array> _domUpdateEvent; // рассылка события обновления стакана заявок
        
        private SafeObservableEvents<Deal> _dealReceiveEvent; // рассылка события прихода сделок

        public Action<Side, int, double, double> _domUpdateRaised;

        #endregion fields

        #region Methods

        /// <summary>
        /// Конструктор
        /// </summary>
        /// <param name="symbol">Инструмент</param>
        /// <param name="shortName">Короткое название</param>
        /// <param name="longName">Длиинное название</param>
        /// <param name="step">Шаг цены</param>
        /// <param name="stepPrice">Стоимость Шага цены</param>
        /// <param name="decimals">Число знаков после запятой</param>
        /// <param name="secExtId">идентификатор биржи</param>
        /// <param name="expiryDate">Дата экспирации</param>
        /// <param name="daysBeforeExpiry">Дней до экспирации</param>
        public Ticker(string symbol, string shortName, string longName,
            double step, double stepPrice, double decimals, string secExtId,
            DateTime expiryDate, double daysBeforeExpiry)
        {
            this._domUpdateEvent = new SafeObservableEvents<Array>();
            this._dealReceiveEvent = new SafeObservableEvents<Deal>();
            
            this._deals = new ObservableCollection<Deal>();
            this._dom = new DepthOfMarket(this);

            this._symbol = symbol;
            this._shortName = shortName;
            this._longName = longName;
            this._secExtId = secExtId;
            this._exchange = null; ;

            this._common = new QuoteCommon(this);

            Update(step, stepPrice, decimals, expiryDate, daysBeforeExpiry);
        }

        /// <summary>
        /// Добавить общую сделку 
        /// </summary>
        /// <param name="price">Цена</param>
        /// <param name="amount">Количество</param>
        /// <param name="dealSide"></param>
        /// <param name="dealId"></param>
        /// <param name="time"></param>
        public void AddDeal(double price, double amount, string dealId, DateTime time)
        {
            Deal newDeal = new Deal(null, this, price, amount, dealId, time);
            this._deals.Add(newDeal);
            this._dealReceiveEvent.Notify(newDeal);
            // DealReceiveRaise(newDeal);
        }

        public void Update(
            double step, double stepprice, double decimals,
            DateTime expiryDate, double daysbeforeexpiry) 
        {
            this._step = step;
            this._stepPrice = stepprice;
            this._decimals = (int)decimals;
            this._expiryDate = expiryDate;
            this._daysBeforeExpiry = daysbeforeexpiry;
        }

        public void UpdateBidsAsks(int index, int numItems,
            double bidPrice, double bidAmount,
            double askPrice, double askAmount)
        {
            this._dom.UpdateBidQuote(index, bidPrice, bidAmount);
            this._dom.UpdateAskQuote(index, askPrice, -askAmount);

            if (index == numItems - 1) this.UpdateDOM();
        }

        public IDisposable SubscribeOnDOMUpdate(Action<Array> actionDOMUpdate)
        {
            return this._domUpdateEvent.Subscribe(actionDOMUpdate);
        }

        public IDisposable SubscribeOnDealReceive(Action<Deal> actionDealRaise)
        {
            return this._dealReceiveEvent.Subscribe(actionDealRaise);
        }

        public double ToMoney(double Punkts)
        {
            return this._stepPrice / this._step * Punkts;
        }

        internal void SetExchange(Exchange exchange)
        {
            this._exchange = exchange;
        }

        private void UpdateDOM()
        {
            Array bids = new object[this._dom.Bids.Count];
            for (int i = 0; i < this._dom.Bids.Count; i++)
            {
                Level level = this._dom.Bids[i];
                Array levelArray = new object[] { level.Price, level.Amount };
                bids.SetValue(levelArray, i);
            }

            Array asks = new object[this._dom.Asks.Count];
            for (int i = 0; i < this._dom.Asks.Count; i++)
            {
                Level level = this._dom.Asks[i];
                Array levelArray = new object[] { level.Price, level.Amount };
                asks.SetValue(levelArray, i);
            }

            this._domUpdateEvent.Notify(new object[] { bids, asks });
        }

        #endregion Methods

        #region Properties

        /// <summary>
        /// Код бумаги
        /// </summary>
        public string Symbol { get { return this._symbol; } }
        
        /// <summary>
        /// Короткое название
        /// </summary>
        public string ShortName { get { return this._shortName; } }
        
        /// <summary>
        /// Пполное наименование инструмента
        /// </summary>
        public string LongName { get { return this._longName; } }
        
        /// <summary>
        /// Шаг цены
        /// </summary>
        public double Step { get { return this._step; } }
        
        /// <summary>
        /// стоимость шага цены
        /// </summary>
        public double StepPrice { get { return this._stepPrice; } }
        
        /// <summary>
        /// число знаков после запятой
        /// </summary>
        public int Decimals { get { return this._decimals; } }
        
        /// <summary>
        /// Идентификатор биржи
        /// </summary>
        public string SecExtId { get { return this._secExtId; } }
        
        /// <summary>
        /// Наименование биржи
        /// </summary>
        public Exchange Exchange { get { return this._exchange; } }
        
        /// <summary>
        /// Дата экспирации
        /// </summary>
        public System.DateTime ExpiryDate { get { return this._expiryDate; } }
        
        /// <summary>
        /// Число дней до экспирации
        /// </summary>
        public double DaysBeforeExpiry { get { return this._daysBeforeExpiry; } }

        /// <summary>
        /// Общая информация по инструменту
        /// </summary>
        public QuoteCommon Common { get { return this._common; } }

        /// <summary>
        /// Общая информация по инструменту
        /// </summary>
        public DepthOfMarket DOM { get { return this._dom; } }


        #endregion Properties

        #region Interface Implementation

        public override string ToString()
        {
            CultureInfo ci = new CultureInfo("en-us");
            return "[" + _exchange.Name + ":" + _symbol + "] Step: " + _step.ToString("G", ci) + " Money: " + ToMoney(1).ToString("G", ci)
                + " Short:" + _shortName + " Expiry: " + _expiryDate.ToShortDateString() + "(" + (int)_daysBeforeExpiry + ")";
        }

        #endregion Interface Implementation
    }
}
