﻿using System;
using System.Collections.Generic;
using System.Collections.Concurrent;
using System.ComponentModel;
using System.Linq;
using System.Text;

using ITIN.CSharp.Types;

namespace ITIN.CSharp.BaseObjects
{   
    /// <summary>
    /// Класс ордера (заявки)
    /// </summary>
    public class TradeOrder : INotifyPropertyChanged
    {
        #region Fields

        /// <summary>
        /// Список отработавших оредов, на этапе отладки необходим для просмотра истории команд по "сбойным" ордерам
        /// В реале, данный список имеет смысл использывать вместо создания нового ордера (см. статический метод GetOrder)
        /// </summary>
        private static ConcurrentQueue<TradeOrder> _freeOrders = new ConcurrentQueue<TradeOrder>();

        /// <summary>
        /// Инструмент
        /// </summary>
        private Ticker _ticker;

        /// <summary>
        /// Стратегия, которая разместила ордер
        /// </summary>
        private IStrategy _strategy;

        /// <summary>
        /// Время размещения(перемещения, отмены ордера)
        /// </summary>
        private Nullable<DateTime> _time;

        /// <summary>
        /// Подписка на освобождение ордера.
        /// Необходима для помещения после снятия или исполнения ордера в очередь "свободных" заявок
                /// </summary>
        private SafeObservableEvents<TradeOrder> _onDisposeOrderEvents;
        
        /// <summary>
        /// Подисчики на событие _onDisposeOrderEvents
        /// </summary>
        private List<IDisposable> _disposeSubscribers;
        
        /// <summary>
        /// Исторя команд или изменений состояния ордера.
        /// Удобно использывать как на начальном этапе разработки торговой бизнес-логики,
        /// так и для отладки работы стратегий
        /// </summary>
        private List<Tuple<OrderCommand, object, DateTime>> _orderHistory;
        
        private double _amount;
        private double _amountFilled;
        private string _exchangeOrderId;
        private int _extId;
        private string _name;
        private string _orderId;
        private double _price;
        private OrderState _state;

        public event PropertyChangedEventHandler PropertyChanged;

        #endregion Fields

        #region Methods
        
        /// <summary>
        /// Конструктор ордера(заявки)
        /// </summary>
        /// <param name="strategy">Стратегия, отправляющая заявку</param>
        /// <param name="account">Счет</param>
        /// <param name="ticker">Инструмен</param>
        /// <param name="price">Цена ордера</param>
        /// <param name="amount">Количество (+ покупка),(- продажа)</param>
        /// <param name="extId">Идентификатор торговой стратегии</param>
        /// <param name="type">Тип ордера</param>
        /// <param name="time">Время размещения ордера</param>
        private TradeOrder(IStrategy strategy, Account account, Ticker ticker, double price, double amount, int extId, OrderType type = OrderType.Limit, Nullable<DateTime> time = null) 
        {
            this._onDisposeOrderEvents = new SafeObservableEvents<TradeOrder>();
            this._disposeSubscribers = new List<IDisposable>();
            this._orderHistory = new List<Tuple<OrderCommand, object, DateTime>>();
        }

        /// <summary>
        /// Добавить команду или изменение по ордеру
        /// </summary>
        /// <param name="strategyDataType">Тип изменения</param>
        /// <param name="commandOption">Необходимые параметры, для восстановления последовательности команд и изменений ордера</param>
        public void AddHistoryCommand(OrderCommand strategyDataType, object commandOption)
        {
            var command = new Tuple<OrderCommand, object, DateTime>(strategyDataType, commandOption, DateTime.Now);
    
            // проверять не будем, так как могут быть случаи когда наша заявка принята системой (pending) и ее можно пытаться отменять
            // эту задачу должен решать поставщик данных, в нашем случае это TradeEngine
            //if (this.LastSendCommand != null) throw new Exception("Предыдущая команда не завершена:" + this.LastSendCommand.ToString());

            if ((strategyDataType == OrderCommand.Send) ||
                (strategyDataType == OrderCommand.Move) ||
                (strategyDataType == OrderCommand.Cancel))
            {
                this.LastSendCommand = command;
            }
            
            this._orderHistory.Add(command);
        }

        /// <summary>
        /// Получить последнюю команду из списка истории ордеров.
        /// Метод потоко не безопасен
        /// </summary>
        /// <returns>Команда(изменение ордера)</returns>
        public Tuple<OrderCommand, object, DateTime> GetLastCommand()
        {
            if (this._orderHistory.Count > 0)
            {
                return this._orderHistory[this._orderHistory.Count - 1];
            }
            return null;
        }

        /// <summary>
        /// Получить последнюю команду типа OrderCommand из списка истории ордеров.
        /// Метод потоко не безопасен
        /// </summary>
        /// <param name="strategyDataType">Тип искомой команды</param>
        /// <returns>Команда(изменение ордера)</returns>
        public Tuple<OrderCommand, object, DateTime> GetLastCommand(OrderCommand strategyDataType)
        {
            if (this._orderHistory.Count > 0)
            {
                Tuple<OrderCommand, object, DateTime> tmpHistoryTuple = null;
                for (int i = this._orderHistory.Count - 1; i >= 0; i--)
                {
                    tmpHistoryTuple = this._orderHistory[i];
                    if (tmpHistoryTuple.Item1 == strategyDataType) return tmpHistoryTuple;
                }
            }
            return null;
        }

        /// <summary>
        /// Омвободить ордер от использывания и поместить в очередь "свободных" ордеров
        /// </summary>
        public void DisposeAndFreeOrder()
        {
            this.State = OrderState.Disposed;
            // уведомим всех подписчиков  о прекращении использывания ордера
            // как правило уведомление исключает ордер из талиц и списков
            this._onDisposeOrderEvents.Notify(this);            

            //Отпишемся от уведомлений и очистим список подписчиков
            foreach (var disposeSubscriber in this._disposeSubscribers)
                disposeSubscriber.Dispose();
            this._disposeSubscribers.Clear();

            // историю не очищаем, рано (проще искат ошибки, имея историю команд ордера)
            // только перед повторным использыванием ордера (в методе InitializeOrder)
            //this._orderHistory.Clear();

            // Оповестим при необходимости стратегию о прекращении использования ордера
            if (OnDisposed != null) OnDisposed(this);

            // поместим ордер в очередь
            TradeOrder._freeOrders.Enqueue(this);
        }

        /// <summary>
        /// Инициализация ордера начальными параметрами и значениями аргументов функции
        /// </summary>
        /// <param name="strategy">Стратегия</param>
        /// <param name="account">Счет</param>
        /// <param name="ticker">Инструмент</param>
        /// <param name="price">Цена</param>
        /// <param name="amount">Количество (+/-)</param>
        /// <param name="extId">Идентификатор торговой стратегии</param>
        /// <param name="type">Тип ордера</param>
        /// <param name="time">Время размещения(изменения) ордера</param>
        public void InitializeOrder(IStrategy strategy, Account account, Ticker ticker, double price, double amount, int extId, OrderType type = OrderType.Limit, Nullable<DateTime> time = null)
        {
            this.OnFailCanceled = null;
            this.OnFailMoved = null;
            this.OnFailOpened = null;
            this.OnPending = null;
            this.OnOpened = null;
            this.OnMoved = null;
            this.OnCanceled = null;
            this.OnDisposed = null;
            this.OnFilled = null;
            this.OnPartialFilled = null;

            this.Account = account;
            this._strategy = strategy;
            this._ticker = ticker;

            this.Price = price;
            this.Amount = amount;
            this.AmountFilled = 0.0;
            this.StopValue = 0.0;
            this.Type = type;
            this.ExtId = extId;
            this.OrderId = "";
            this.ExchangeOrderId = "0";
            // по умолчанию, считаем, что ордер не GTD
            this._time = time;

            this.LastSendCommand = null;

            this._orderHistory.Clear();
        }

        /// <summary>
        /// Метод подписки на рассылку "удаления"(снятия) ордера
        /// </summary>
        /// <param name="onDisposeOrder">Обработчик</param>
        public void SubscribeOnDisposeOrder(Action<TradeOrder> onDisposeOrder)
        {
            this._disposeSubscribers.Add(this._onDisposeOrderEvents.Subscribe(onDisposeOrder));
        }

        /// <summary>
        /// Получить ордер (создать или выбрать из очереди "свободных" ордеров) 
        /// </summary>
        /// <param name="strategy">Стратегия</param>
        /// <param name="account">Счет</param>
        /// <param name="ticker">Инструмент</param>
        /// <param name="price">Цена</param>
        /// <param name="amount">Количество (+/-)</param>
        /// <param name="extId">Идентификатор торговой стратегии</param>
        /// <param name="type">Тип ордера</param>
        /// <param name="time">Время размещения(изменения) ордера</param>
        /// <returns>Экземпляр заявки</returns>
        public static TradeOrder GetInitializedOrder(IStrategy strategy, Account account, Ticker ticker, double price, double amount, int extId, OrderType type = OrderType.Limit, Nullable<DateTime> time = null)
        {
            TradeOrder tmpOrder = null;
            //тестовый вариант
            tmpOrder = new TradeOrder(strategy, account, ticker, price, amount, extId, type, time);
            
            // "боевой" вариант
            /*if (TradeOrder._freeOrders.Count == 0)
            {
                tmpOrder = new TradeOrder(strategy, account, ticker, price, amount, extId, type, time);
            }
            else
            {
                TradeOrder._freeOrders.TryDequeue(out tmpOrder);
            }*/

            tmpOrder.InitializeOrder(strategy, account, ticker, price, amount, extId, type, time);
            return tmpOrder;
        }

        /// <summary>
        /// Сбросить последнюю команду управления ордером
        /// </summary>
        public void ResetLastSendCommand()
        {
            this.LastSendCommand = null;
        }

        #endregion Methods

        #region Properties

        /// <summary>
        /// Счет
        /// </summary>
        public Account Account { get; private set; }

        /// <summary>
        /// Объем
        /// </summary>
        public double Amount
        {
            get { return this._amount; }
            
            set
            {
                this._amount = value;
                OnPropertyChanged("Account");
            }
        }

        /// <summary>
        /// Исполненный объем
        /// </summary>
        public double AmountFilled
        {
            get { return this._amountFilled; }
            
            set
            {
                this._amountFilled = value;
                OnPropertyChanged("AmountFilled");
            }
        }
        
        /// <summary>
        /// Остаток в заявке
        /// </summary>
        public double AmountRest { get { return this.Amount - this.AmountFilled; } }

        /// <summary>
        /// Идентификатор биржи
        /// </summary>
        public string ExchangeOrderId
        {
            get { return this._exchangeOrderId; }
            
            set
            {
                this._exchangeOrderId = value;
                OnPropertyChanged("ExchangeOrderId");
            }
        }
        
        /// <summary>
        /// Идентификатор торговой стратегии
        /// </summary>
        public int ExtId
        {
            get { return this._extId; }
            
            set
            {
                this._extId = value;
                OnPropertyChanged("ExchangeOrderId");
            }
        }

        /// <summary>
        /// Последняя команда управления ордером
        /// Должна лишь принимать значения Send, Move Cancel и т.п.
        /// </summary>
        public Tuple<OrderCommand, object, DateTime> LastSendCommand { get; private set; }

        /// <summary>
        /// Имя ордера
        /// </summary>
        public string Name
        {
            get { return this._name; }
            
            set
            {
                this._name = value;
                OnPropertyChanged("Name");
            }
        }

        /// <summary>
        /// Инентификатор прокера, или же комментарий к ордеру.
        /// Особенностью данного свойства являеться следующее условие: OrderId - величина постоянная для 1 ордера
        /// </summary>
        public string OrderId
        {
            get { return this._orderId; }
            
            set
            {
                this._orderId = value;
                OnPropertyChanged("OrderId");
            }
        }

        /// <summary>
        /// Событие, уведомляющее о не удачной отмене заявки
        /// </summary>
        public Action<TradeOrder> OnFailCanceled { get; set; }
        
        /// <summary>
        /// Событие, уведомляющее о не удачной перестановки(перемещении) заявки
        /// </summary>
        public Action<TradeOrder, double, double> OnFailMoved { get; set; }
        
        /// <summary>
        /// Событие, уведомляющее о не удачной отправки(выставлении) заявки
        /// </summary>
        public Action<TradeOrder, string> OnFailOpened { get; set; }
        
        /// <summary>
        /// Событие о принятии заявки сервером
        /// </summary>
        public Action<TradeOrder> OnPending { get; set; }

        /// <summary>
        /// Событие о принятии заявки торговой площадкой(биржей)
        /// </summary>
        public Action<TradeOrder> OnOpened { get; set; }
        
        /// <summary>
        /// Событие удачного перемещения заявки
        /// </summary>
        public Action<TradeOrder, double, double> OnMoved { get; set; }
        
        /// <summary>
        /// Событие об удачной отмене(снятии) заявки
        /// </summary>
        public Action<TradeOrder> OnCanceled { get; set; }
        
        /// <summary>
        /// Событие освобождения ордера
        /// </summary>
        public Action<TradeOrder> OnDisposed { get; set; }
        
        /// <summary>
        /// Событие о полном исполнении заявки
        /// </summary>
        public Action<TradeOrder> OnFilled { get; set; }

        /// Событие о частичном исполнении заявки
        public Action<TradeOrder, double> OnPartialFilled { get; set; }

        /// <summary>
        /// Цена заявки
        /// </summary>
        public double Price
        {
            get { return this._price; }
            
            set
            {
                this._price = value;
                OnPropertyChanged("Price");
            }
        }

        /// <summary>
        /// Направление заявки, определяеться  положительныи или отрицательным значением 
        /// </summary>
        public Side Side
        {
            get
            {
                if (this.Amount == 0.0)
                {
                    throw new Exception("Ордер не можен иметь нулевой объем!");
                    //return (Side)Math.Sign(AmountRest);
                }
                else 
                {
                    return (Side)Math.Sign(Amount);
                }
            }
        }
        
        /// <summary>
        /// Состояние аявки
        /// </summary>
        public OrderState State
        {
            get { return this._state; }
            
            set
            {
                this._state = value;
                OnPropertyChanged("State");
            }
        }

        /// <summary>
        /// Стратегия
        /// </summary>
        public IStrategy Strategy { get { return this._strategy; } }

        /// <summary>
        /// "Стоп" значение
        /// </summary>
        public double StopValue { get; private set; }

        /// <summary>
        /// Инструмент
        /// </summary>
        public Ticker Ticker { get { return this._ticker; } }

        /// <summary>
        /// Время размещения (изменения) заявки
        /// </summary>
        public Nullable<DateTime> Time
        {
            get { return this._time; }
            set
            {
                this._time = value;
                this.OnPropertyChanged("Time");
            }
        }

        /// <summary>
        /// Ттип заявки
        /// </summary>
        public OrderType Type { get; private set; }

        #endregion Properties

        #region Handlers
        protected void OnPropertyChanged(string name)
        {
            PropertyChangedEventHandler handler = PropertyChanged;
            if (handler != null)
            {
                handler(this, new PropertyChangedEventArgs(name));
            }
        }
        #endregion Handlers

        #region Interface Implementation

        public override string ToString()
        {
            string timeString = " \t ";
            if (_time != null)
            {
                timeString = _time.Value.ToLongTimeString();
            }
            return timeString + Name + "[" + this.ExtId.ToString() + ":" + this.Ticker.Symbol + "] " + Price.ToString() + " \t Amount= " + Amount.ToString() + " \t AmountRest= " + AmountRest.ToString()
                + " \t State: " + this.State.ToString() + " \t Order ID" + OrderId + " \t Exchange ID" + ExchangeOrderId;
        }

        #endregion Interface Implementation
    }
}
