﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using ITIN.CSharp.Types;

namespace ITIN.CSharp.BaseObjects
{
    /// <summary>
    /// Таблица ордеров стратегии
    /// </summary>
    public class TradeOrdersTable
    {
        #region Fields
        
        /// <summary>
        /// Ордера с поиском по идентификатору стратегии
        /// </summary>
        private SafeObservableDictionary<string, TradeOrder> _orders;
        
        /// <summary>
        /// Ордера с поиском по идентификатору сервера ил комментарию (ключ - постоянная величина в рамках жизни одного ордера)
        /// </summary>
        private SafeObservableDictionary<int, TradeOrder> _ordersExt;

        #endregion Fields

        #region Methods

        /// <summary>
        /// Конструктор
        /// </summary>
        public TradeOrdersTable()
        {
            this._orders = new SafeObservableDictionary<string, TradeOrder>();
            this._ordersExt = new SafeObservableDictionary<int, TradeOrder>();
        }

        /// <summary>
        /// Добавить ордер в список ордеров стратегии
        /// </summary>
        /// <param name="extId">Идентификатор стратеги</param>
        /// <param name="order">Помещаемый ордер</param>
        public void AddOrder(int extId, TradeOrder order)
        {
            if (order.State == OrderState.Disposed) return;

            if (this._ordersExt.TryAdd(extId, order) == true)
            {
                // при прекращении испольывания ордера, исключим ордер из списка
                order.SubscribeOnDisposeOrder((orderToDispose) =>
                {
                    this.RemoveOrder(orderToDispose.ExtId);
                });
            }
            //else System.Diagnostics.Debug.WriteLine("Ордер с ключом extId=" + extId.ToString() + " уже присутсвтует.");
        }

        /// <summary>
        /// Добавить ордер в список ордеров стратегии
        /// </summary>
        /// <param name="order">Помещаемый ордер</param>
        public void AddOrder(TradeOrder order)
        {
            if (order.State == OrderState.Disposed) return;

            // если имя заявки указываетсья в качестве ключа, то мы добавляем ордер в списки стратегий
            if (this._orders.TryAdd(order.Name, order) == true)
            {
                // при прекращении испольывания ордера, исключим ордер из списка
                order.SubscribeOnDisposeOrder((orderToDispose) =>
                {
                    this.RemoveOrder(orderToDispose.Name);
                });
            }
            //else System.Diagnostics.Debug.WriteLine("Ордер с ключом Name=" + order.Name + " уже присутсвтует.");
        }

        /// <summary>
        /// Добавить ордер в список ордеров сервера
        /// </summary>
        /// <param name="orderId">Идентификатор сервера (комментарий)</param>
        /// <param name="order">Помещаемый ордер</param>
        public void AddOrder(string orderId, TradeOrder order)
        {
            if (order.State == OrderState.Disposed) return;

            if (this._orders.TryAdd(orderId, order) == true)
            {
                // при прекращении испольывания ордера, исключим ордер из списка
                order.SubscribeOnDisposeOrder((orderToDispose) =>
                {
                    this.RemoveOrder(orderToDispose.OrderId);
                });
            }
            //else System.Diagnostics.Debug.WriteLine("Ордер с ключом orderId=" + orderId.ToString() + " уже присутсвтует.");
        }

        /// <summary>
        /// Получитьордер по ключу из списка ордеров стратегии
        /// </summary>
        /// <param name="extId">Ключ - идентификатор заявки стратегии</param>
        /// <returns>Искомый ордер или null в случае отсутствия ключа</returns>
        public TradeOrder GetOrder(int extId)
        {
            TradeOrder tmpOrder = null;
            if (this._ordersExt.TryGetValue(extId, out tmpOrder) == true)
            {
                return tmpOrder;
            }
            return null;
            //throw new Exception("Не найдена заявка extId=" + extId.ToString());
        }

        /// <summary>
        /// Получитьордер по ключу из списка ордеров сервера
        /// </summary>
        /// <param name="orderId">Ключ - идентификатор сервера (комментарий)</param>
        /// <returns>Искомый ордер или null в случае отсутствия ключа</returns>
        public TradeOrder GetOrder(string orderId)
        {
            TradeOrder tmpOrder = null;
            if (this._orders.TryGetValue(orderId, out tmpOrder) == true)
            {
                return tmpOrder;
            }
            return null;
            //throw new Exception("Не найдена заявка extId=" + orderId.ToString());
        }

        /// <summary>
        /// Получить список всех ордеров с внешним фильтром
        /// </summary>
        /// <returns>Список ордеров</returns>
        public List<TradeOrder> GetFilteredOrders(Func<TradeOrder, bool> filter)
        {
            return new List<TradeOrder>(
                    from order in this._orders
                    where (filter(order.Value) == true)
                    select order.Value);
        }

        /// <summary>
        /// Получить список всех ордеров сервера
        /// </summary>
        /// <returns>Список ордеров</returns>
        public List<TradeOrder> GetOrders()
        {
            return new List<TradeOrder>(
                    from order in this._orders
                    select order.Value);
        }

        /// <summary>
        /// Получить список всех ордеров стратегии
        /// </summary>
        /// <returns>Список ордеров</returns>
        public List<TradeOrder> GetOrdersExt()
        {
            return new List<TradeOrder>(
                    from order in this._ordersExt
                    select order.Value);
        }

        /// <summary>
        /// Получить потокобезопасную коллекцию-словарь ордеров сервера
        /// </summary>
        /// <returns>Ссылка на потокобезопасную коллекцию-словарь</returns>
        public SafeObservableDictionary<string, TradeOrder> GetOrdersDictionary()
        {
            return this._orders;
        
        }

        /// <summary>
        /// Получить потокобезопасную коллекцию-словарь ордеров стратегии
        /// </summary>
        /// <returns>Ссылка на потокобезопасную коллекцию-словарь</returns>
        public SafeObservableDictionary<int, TradeOrder> GetOrdersExtDictionary()
        {
            return this._ordersExt;
        }

        /// <summary>
        /// Удалить ордер стратегии
        /// </summary>
        /// <param name="extId">Ключ идентификатор ордера стратегии</param>
        private void RemoveOrder(int extId)
        {
            TradeOrder tmpOrder = null;
            this._ordersExt.TryRemove(extId, out tmpOrder);
            /*if (this._ordersExt.TryRemove(extId, out tmpOrder) == false)
            {
                throw new Exception("Ордер extId=" + extId.ToString() + "для удаления не найден ");
            }*/
        }

        /// <summary>
        /// Удалить ордер стратегии
        /// </summary>
        /// <param name="extId">Ключ идентификатор ордера стратегии</param>
        private void RemoveOrder(string orderId)
        {
            TradeOrder tmpOrder = null;
            this._orders.TryRemove(orderId, out tmpOrder);
            /*if (this._orders.TryRemove(orderId, out tmpOrder) == false)
            {
                throw new Exception("Ордер extId=" + orderId.ToString() + "для удаления не найден ");                
            }*/
        }

        #endregion Methods

        #region Properties

        

        #endregion Properties
    }
}
