﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

using ITIN.CSharp.BaseObjects;
using ITIN.CSharp.Types;

namespace ITIN.CSharp.Strategy
{
    /// <summary>
    /// Генератор идентификаторов заявок стратегий
    /// В данном случае выбрали простой вариант
    /// 
    /// Возможно использывание 7 старших бит из 32 бит(4 байта int) 
    /// Получим возможность кодировать 127 стратегий , слева остаеться 25 бит (33 554 431 ордеров в рамках одной стратегии за торговую сессию/день)
    public class ExtIdGenerator
    {
        private static ExtIdGenerator _generator = new ExtIdGenerator();
        private DateTime _startTime = DateTime.Now;
        private volatile int _index = DateTime.Now.Hour * 10000000 + DateTime.Now.Minute * 1000000;

        /// <summary>
        /// Получить идентификатор стратегии
        /// </summary>
        /// <returns></returns>
        public static int GetExtId()
        {
            return Interlocked.Increment(ref _generator._index);
        }

        /// <summary>
        /// Получить время старта
        /// Необходимо для поиска ордера, по котороому пришла сделка.
        /// На момент прихолда сделки биржевой идентификатор мождет быть неизвестен.
        /// </summary>
        /// <returns></returns>
        public static DateTime StartTime { get { return _generator._startTime; } }
    }


    /// <summary>
    /// Базовый класс стратегий
    /// Позволяет выполнять элементарные операции по выставлению перемещению и отмене ордеров, а также обработки событий изменения ордера
    /// </summary>
    public class BaseStrategy : IStrategy, INotifyPropertyChanged

    {
        #region Fields
        
        /// <summary>
        /// Указывает запущена(1) или остановлена(0) стратегия
        /// </summary>
        private int _isStarted;
        
        /// <summary>
        /// Очередь заданий стратегии
        /// </summary>
        private ConcurrentQueue<Array> _strategyActionsQueue;
        
        /// <summary>
        /// Событие о поступлении нового задания
        /// </summary>
        private AutoResetEvent _ARE;

        /// <summary>
        /// Событие о изменение стакана, заполняеться стратегиями потомками
        /// </summary>
        private Action<Array, Array> _onStrategyDOMUpdateEvent;
        
        /// <summary>
        /// /// Событие о получении сделки по инструменту, заполняеться стратегиями потомками
        /// </summary>
        private Action<Deal> _onStrategyDealReceiveEvent;
        
        /// <summary>
        /// Событие Таймера, заполняеться стратегиями потомками
        /// </summary>
        private Action _onTimerEvent;
        
        /// <summary>
        /// Ссылка на торговый движок для взаимодействия
        /// </summary>
        private IEngine _engineData;

        /// <summary>
        /// Наименование стратегии
        /// </summary>
        private string _name;

        /// <summary>
        /// Таблица ордеров
        /// </summary>
        private TradeOrdersTable _orders;

        /// <summary>
        /// Задача по исполнению очереди заданий. Необходима например для определения момента остановки (прекращения) работы стратегии
        /// </summary>
        private Task _strategyTask;

        /// <summary>
        /// Очередь внешних подписок. 
        /// Используеться для упрощения "отписок" от рассылок при остановке стратегии
        /// </summary>
        private ConcurrentQueue<IDisposable> _strategyExternalSubscribers;
        
        public event PropertyChangedEventHandler PropertyChanged;

        #endregion Fields

        #region Methods

        /// <summary>
        /// Конструктор
        /// </summary>
        /// <param name="engineData">Указатель на ядро </param>
        /// <param name="name">Наименование стратегии</param>
        public BaseStrategy(IEngine engineData, string name)
        {
            this._orders = new TradeOrdersTable();
            this._name = name;
            this._isStarted = 0;
            this._engineData = engineData;
            this._strategyActionsQueue = new ConcurrentQueue<Array>();
            this._strategyExternalSubscribers = new ConcurrentQueue<IDisposable>();
            this._ARE = new AutoResetEvent(false);
            this._onStrategyDOMUpdateEvent = null;
            this._onStrategyDealReceiveEvent = null;
            this._onTimerEvent = null;            
        }

        /// <summary>
        /// Поместить в очередь стратегии SGH задание на исполнение в потоке стратегии
        /// </summary>
        /// <param name="action">Действие, котороенеобходимо выполнить в потоке стратегии</param>
        public void SendStrategyAction(Action action)
        {
            if (action != null)
            {
                Array message = new object[] { StrategyDataType.SGH, action };
                this.SendStrategyData(message);
            }
        }

        /// <summary>
        /// Отправить событие в виде массива данных
        /// </summary>
        /// <param name="data"></param>
        public void SendStrategyData(Array data)
        {
            if (data.Length > 0)
            {
                this._strategyActionsQueue.Enqueue(data);
                this._ARE.Set();
            }
        }

        /// <summary>
        /// Запуск стратегии
        /// </summary>
        public void Start()
        {
            // потокобезопасно проверим, если стратегия уже запущена, выходим
            int prevValue = Interlocked.CompareExchange(ref this._isStarted, 1, 0);
            if (prevValue == 1) return;

            OnPropertyChanged("IsStarted");

            this._strategyTask = Task.Factory.StartNew(() =>
                {
                    
                    OnStarted();

                    Thread.Sleep(3000);
                    //дождемся данных по счету
                    this.StrategyWorker();
                }, TaskCreationOptions.LongRunning)
                .ContinueWith((task) => System.Diagnostics.Debug.WriteLine("Исключение в стратегии " + this.Name));
        }

        /// <summary>
        /// Останов стратегии
        /// </summary>
        /// <returns>Текущая задача обработки заданий стратегии (необходима для ожидания останова выполнения заданий стратегии)</returns>
        public Task Stop()
        {
            ///отпишемся от всех рассылок
            IDisposable subscriber = null;
            while (this._strategyExternalSubscribers.Count > 0)
            {   
                if (this._strategyExternalSubscribers.TryDequeue(out subscriber) == false) continue;
                subscriber.Dispose();
            }
            
            // учтем факт, что с большой долей вероятности, данный метод будет вызываться из другого потока
            OnStopped();

            return this._strategyTask;
        }

        /// <summary>
        /// Подписка на изменение стакана по инструменту
        /// </summary>
        /// <param name="tradeTicker">Инструмент</param>
        /// <param name="onDOMUpdateRaise">Внешний обработчик</param>        
        public void SubscribeOnDOMUpdate(Ticker tradeTicker, Action<Array, Array> onDOMUpdateRaise)
        {
            if (tradeTicker == null) return;
            IDisposable onDOMUpdateEvent = tradeTicker.SubscribeOnDOMUpdate((arraySnapshots) =>
                {
                    Array messageData = new object[] { StrategyDataType.DOMUpdate, arraySnapshots };
                    
                    this.SendStrategyData(messageData);
                });

            // установим новый обработчик
            this._onStrategyDOMUpdateEvent = onDOMUpdateRaise;

            // не потеряем подписчика, да последующей "отписки" от события
            this._strategyExternalSubscribers.Enqueue(onDOMUpdateEvent);
        }

        /// <summary>
        /// Подписка получение сделок
        /// Необходимо не потерять IDisposable результат-указатель на "отписку" от события 
        /// </summary>
        /// <param name="tradeTicker">Инструмент</param>
        /// <param name="onDealRaise">Внешний обработчик</param>
        public void SubscribeOnDealReceive(Ticker tradeTicker, Action<Deal> onDealRaise)
        {
            if (tradeTicker == null) return;
            IDisposable onDealReceiver = tradeTicker.SubscribeOnDealReceive((deal) =>
            {
                Array messageData = new object[] { StrategyDataType.Deal, deal };

                this.SendStrategyData(messageData);
            });

            // установим новый обработчик
            this._onStrategyDealReceiveEvent = onDealRaise;

            // не потеряем подписчика, да последующей "отписки" от события
            this._strategyExternalSubscribers.Enqueue(onDealReceiver);
        }


        /// <summary>
        /// Метод разбора очереди заданий стратегии (исполняеться в Task)
        /// </summary>
        private void StrategyWorker()
        {
            Action onDOMUpdateRaise = null;
            Action onDealRaise = null;
            Action onTimerRaise = null;

            // сюда можно добавить переменную или свойство стратегии для события OnTimer, при условии,
            // что велечина не более 1000 мс. При значениях более 1000 мс, рекомендуеться создать отдельный 
            // поток для генерации сообщения OnTimer, или найти другие способы.
            int onTimerMessageWait = 300;

            while ((this._isStarted == 1) || (this._strategyActionsQueue.Count != 0)) // выберем все задания
            {
                if (this._strategyActionsQueue.Count == 0)
                {
                    
                    bool isReceiveSignal = this._ARE.WaitOne(onTimerMessageWait);
                    if (isReceiveSignal == false)
                    {
                        // простой способ генерации событий Таймера
                        this.SendStrategyData(new object[] { StrategyDataType.OnTimer });
                    }
                    continue;
                }


                Array data = null;
                if (this._strategyActionsQueue.TryDequeue(out data) == false) continue;

                // Данные распологаються по определенным индексам (протокол см. в TradeEngine)
                StrategyDataType dataType = (StrategyDataType)(data.GetValue(0));

                switch (dataType)                    
                {
                    case StrategyDataType.DOMUpdate:
                        if (onDOMUpdateRaise == null)
                        {
                            onDOMUpdateRaise = () =>
                            {
                                Array bids = (Array)((Array)data.GetValue(1)).GetValue(0);
                                Array asks = (Array)((Array)data.GetValue(1)).GetValue(1);
                                this._onStrategyDOMUpdateEvent(bids, asks);
                            };
                        }

                        break;

                    case StrategyDataType.OnTimer:
                        if (onTimerRaise == null)
                        {
                            onTimerRaise = () =>
                            {
                                //if (this._onTimerEvent != null) this._onTimerEvent();
                                this._onStrategyDOMUpdateEvent(null, null);
                            };
                        }

                        break;

                    case StrategyDataType.Deal:
                        if (onDealRaise == null)
                        {
                            Deal deal = (Deal)data.GetValue(1);
                            if (this._onStrategyDealReceiveEvent != null) this._onStrategyDealReceiveEvent(deal);
                        }
                        break;

                    case StrategyDataType.SGH:
                        // выполнить 
                        Action strategyAction = (Action)data.GetValue(1);
                        
                        strategyAction();
                        break;
                }

                // если стратегия остановлена, то не имеет смысла запускать методы, генерирующие новые заявки
                if (this._isStarted != 1) continue;
                
                //обработаем событие апдейтап стаканов стратегии
                if (onDOMUpdateRaise != null) onDOMUpdateRaise();
                if (onTimerRaise != null) onTimerRaise();

                //сбросим их
                onDOMUpdateRaise = null;
                onTimerRaise = null;
            }

            System.Diagnostics.Debug.WriteLine("Задача Стратегии " + this.Name + " - остановлена");
        }

        #endregion Methods

        #region Properties

        /// <summary>
        /// Ссылка на ядро торговуй системы
        /// </summary>
        public IEngine Engine { get { return this._engineData; } }

        /// <summary>
        /// Имя стратегии
        /// </summary>
        public string Name { get { return this._name; } }

        /// <summary>
        /// Возвращает, запущена стратегия или нет
        /// </summary>
        public bool IsStarted { get { return this._isStarted == 1; } }


        /// <summary>
        /// Таблица заявок
        /// </summary>
        public TradeOrdersTable Orders { get { return this._orders; } }

        #endregion Properties

        #region Handlers
        protected void OnPropertyChanged(string name)
        {
            PropertyChangedEventHandler handler = PropertyChanged;
            if (handler != null)
            {
                handler(this, new PropertyChangedEventArgs(name));
            }
        }
        #endregion Handlers

        #region Interface implementation

        /// <summary>
        /// Возвращает позицию стратегии по указанному инструменту
        /// </summary>
        /// <param name="ticker"></param>
        /// <returns>Текущая позиция</returns>
        public double GetPosition(Ticker ticker)
        {
            return 0.0;
        }

        /// <summary>
        /// Строка, перечисляющая все позиции
        /// </summary>
        public string Positions { get { return ""; } } 

        /// <summary>
        /// Виртуальный метод запуска стратегии.
        /// Вызываеться из потока задачи (Task) перед началом обработки очереди заданий
        /// </summary>
        protected virtual void OnStarted() { }

        /// <summary>
        /// Виртуальный метод остановки стратегии.
        /// Вызываеться перед началом остановки
        /// </summary>
        protected virtual void OnStopped()
        {
            // потокобезопасно проверим, если стратегия уже остановлена, выходим
            int prevValue = Interlocked.CompareExchange(ref this._isStarted, 0, 1);
            if (prevValue == 0) return;
            
            OnPropertyChanged("IsStarted");
        }

        /// <summary>
        /// Орпавка ордера
        /// </summary>
        /// <param name="account">Счет</param>
        /// <param name="ticker">Инструмент</param>
        /// <param name="price">Цена</param>
        /// <param name="amount">Количество</param>
        /// <param name="orderName">Имя ордера</param>
        /// <returns>Экземпляр заявки</returns>
        public TradeOrder SendOrder(Account account, Ticker ticker, double price, double amount, string orderName = null)
        {
            int extId = ExtIdGenerator.GetExtId();
            string sendOrderName = extId.ToString();
            if (orderName != null)
            {
                sendOrderName = orderName;
            }
            TradeOrder newOrder = this._orders.GetOrder(sendOrderName);
            if (newOrder != null) return null;

            newOrder = TradeOrder.GetInitializedOrder(this, account, ticker, price, amount, extId);
            newOrder.Name = sendOrderName;
            newOrder.State = OrderState.Sending;
            this._orders.AddOrder(newOrder);

            newOrder.AddHistoryCommand(OrderCommand.Send, this.Name + " - " + newOrder.Name);
            this._engineData.SendOrder(newOrder);

            return newOrder;
        }

        /// <summary>
        /// Выставляет или перемещает ордер, при изменении цены на величину minMovingDeltaPrice
        /// </summary>
        /// <param name="account">Счёт</param>
        /// <param name="ticker">Инструмент</param>
        /// <param name="price">Цена</param>
        /// <param name="amount">Количество</param>
        /// <param name="minMovingDeltaPrice">изменение цены, при котором необходимо переместить заявку</param>
        /// <param name="orderName">Имя ордера</param>
        /// <returns>Экземпляр заявки</returns>
        public TradeOrder SendOrMoveOrder(Account account, Ticker ticker, double price, double amount, double minMovingDeltaPrice, string orderName = null)
        {
            TradeOrder tmpOrder = null;
            if (orderName != null)
            {
                tmpOrder = this._orders.GetOrder(orderName);
            }
            if (tmpOrder == null) return SendOrder(account, ticker, price, amount, orderName);

            // определим текущую разницу цен 
            // и сверим с допустимой величиной, при которой заявка будет перемещатсья
            // если разница цен превысила допустимое значение 

            double tmpDeltaPrice = Math.Abs(tmpOrder.Price - price);
            if (tmpDeltaPrice >= minMovingDeltaPrice)
            {
                MoveOrder(tmpOrder, price);
                // необходимо понять, необходимо ли модифицировать обработчики событий ордера
                return tmpOrder;
            }
            
            return null;
        }

        /// <summary>
        /// Снятие заявки. При ожидании снятия заявки, необходимо учесть что некоторые обработчики изменения ордера подменяються
        /// </summary>
        /// <param name="tmpOrder">Заявка</param>
        /// <param name="waitOpenOrderToCancel">Ждать снятия заявки</param>
        public void CancelOrder(TradeOrder tmpOrder, bool waitOpenOrderToCancel = false)
        {
            if ((tmpOrder.State == OrderState.Pending)
              ||(tmpOrder.State == OrderState.Open))
            {
                tmpOrder.State = OrderState.Canceling;
                tmpOrder.AddHistoryCommand(OrderCommand.Cancel, tmpOrder.OrderId);
                this._engineData.CancelOrder(tmpOrder);
                return;
            }

            if (waitOpenOrderToCancel == true)
            {
                // принудительно сброссим все ожидающие ответов события на отмену ордера
                tmpOrder.OnMoved = (ord, oldPrice, newPrice) => CancelOrder(ord);
                tmpOrder.OnPending = (ord) => CancelOrder(ord);
                tmpOrder.OnOpened = (ord) => CancelOrder(ord);
                //tmpOrder.OnFilled = (ord) => CancelOrder(ord);
            }
        }

        /// <summary>
        /// Снятие заявки. При ожидании снятия заявки, необходимо учесть что некоторые обработчики изменения ордера подменяються
        /// </summary>
        /// <param name="orderName">Имя ордера</param>
        /// <param name="waitOpenOrderToCancel">Ждать снятия заявки</param>
        public void CancelOrder(string orderName, bool waitOpenOrderToCancel = false)
        {
            TradeOrder tmpOrder = this._orders.GetOrder(orderName);
            if (tmpOrder == null) return;
            CancelOrder(tmpOrder);
        }

        /// <summary>
        /// Отменить все заявки стратегии
        /// </summary>
        /// <param name="actionAfterCancelAllOrders">Действие, которое необходимо выполнить после отмены всех заявок</param>
        public void CancelAllOrder(Action actionAfterCancelAllOrders = null)
        {
            var orders = this._orders.GetOrders();
            foreach (var order in orders)
            {
                CancelOrder(order, true);
            }
            if (orders.Count != 0)
            {
                Action strategyAction = () =>
                {
                    CancelAllOrder(actionAfterCancelAllOrders);
                };
                this.SendStrategyData(new object[] { StrategyDataType.SGH, strategyAction });
            }
            else
            {
                if (actionAfterCancelAllOrders != null) actionAfterCancelAllOrders();
            }
        }

        public void ClosePosition(Ticker ticker, Action closePosition, Action onClosedPosition = null)
        {
            var tmpTicker  = ticker;
            var orders = this._orders.GetOrders();
            var position = GetPosition(tmpTicker);

            if ((position == 0.0) && (orders.Count == 0.0))
            {
                if (onClosedPosition != null) onClosedPosition();
            }

            if (orders.Count != 0)
            {
                Action strategyAction = () =>
                {
                    ClosePosition(tmpTicker, closePosition, onClosedPosition);
                };
                this.SendStrategyData(new object[] { StrategyDataType.SGH, strategyAction });
            }
            else
            {
                
            }
        }        

        /// <summary>
        /// Переместить ордер.
        /// </summary>
        /// <param name="tmpOrder">Перемещяемый ордер</param>
        /// <param name="newPrice">Новое значение цены</param>
        public void MoveOrder(TradeOrder tmpOrder, double newPrice)
        {
            if (//(tmpOrder.State != OrderState.Pending) &&
                (tmpOrder.State != OrderState.Open)) return;

            // нет смысла перемещать заявку по одной и той же цене
            if (tmpOrder.Price == newPrice) return;

            tmpOrder.State = OrderState.Moving;
            tmpOrder.AddHistoryCommand(OrderCommand.Move, newPrice);
            this._engineData.MoveOrder(tmpOrder, newPrice);
        }

        /// <summary>
        /// Переместить ордер.
        /// </summary>
        /// <param name="orderName">Имя перемещяемого ордера</param>
        /// <param name="newPrice">Новое значение цены</param>
        public void MoveOrder(string orderName, double newPrice)
        {
            TradeOrder tmpOrder = this._orders.GetOrder(orderName);
            if (tmpOrder != null)
            {
                MoveOrder(tmpOrder, newPrice);
            }
            //else
            //{
            //    throw new Exception("Стратегия " + this.Name + " не содержит ордера с именем " + tmpOrder.Name);
            //}
        }

        #endregion Interface implementation

    }
}
