﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Collections.Specialized;
using System.ComponentModel;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Threading;

//using StClientLib; // SmartCOM 2
using SmartCOM3Lib; // SmartCOM 3

using ITIN.CSharp;
using ITIN.CSharp.BaseObjects;
using ITIN.CSharp.SmartCOM;
using ITIN.CSharp.Strategy;
using ITIN.CSharp.Types;

namespace ITIN.CSharp.SmartCOM
{
    /// <summary>
    /// класс движка алгоритмической торговли с помощью SmartCOM
    /// </summary>
    public class TradeEngine : IEngine
    {
        #region Fields

        /// <summary>
        /// COM Сервер 
        /// </summary>
        private StServerClass _smartCOMServer;

        /// <summary>
        /// Адреса и параметры возможных серверов
        /// </summary>
        private SmartCOMServerCollection _servers;

        /// <summary>
        /// Объект уведомлений о состоянии соединения
        /// </summary>
        private ServerConnectionStateNotifier<SmartCOMServerConnectionParametrs> _connectionNotifier;

        /// <summary>
        /// Все инструменты
        /// </summary>
        private ConcurrentDictionary<string, Ticker> _instruments;

        /// <summary>
        /// Инструменты с группировкой по биржам
        /// </summary>
        private ConcurrentDictionary<string, Exchange> _excangeInstruments;

        /// <summary>
        /// Список доступных торговых счетов
        /// </summary>
        private ConcurrentDictionary<string, Account> _accounts;

        /// <summary>
        /// Созданные стратегии
        /// </summary>
        private SafeObservableDictionary<string, IStrategy> _strategies;

        /// <summary>
        /// Собыьтие прихода инструмента RI
        /// </summary>
        private Action<Ticker> _tickerReceivedRaise;

        /// <summary>
        /// Поток Ядра
        /// </summary>
        private EngineWorkerThread _workerThread;

        /// <summary>
        /// Таблица заявок для поиска ордеров
        /// </summary>
        private TradeOrdersTable _ownOrdersTable;
        
        /// <summary>
        /// Запущен ли тест метод
        /// </summary>
        private volatile bool _isTestRunning;

        /// <summary>
        /// Проверка на принудительный разрыв или ошибку сети
        /// </summary>
        private volatile bool _isUserDisconnect;
        
        /// <summary>
        /// Параметры выбранного подключения к серверу
        /// </summary>
        private SmartCOMServerConnectionParametrs _currentServerConnectionParametrs;

        #endregion Fields


        // Для отображения стакана
        public INotifyDOMUpdated _notifierDOMUpdate;
        public ObservableCollection<Level> _visualBids;
        public ObservableCollection<Level> _visualAsks;

        #region Methods

        /// <summary>
        /// Конструктор
        /// </summary>
        public TradeEngine()
        {
            this._connectionNotifier = new ServerConnectionStateNotifier<SmartCOMServerConnectionParametrs>();
            this._tickerReceivedRaise = null;
            this._instruments = new ConcurrentDictionary<string, Ticker>();
            this._excangeInstruments = new ConcurrentDictionary<string, Exchange>();
            this._accounts = new ConcurrentDictionary<string, Account>();
            this._strategies = new SafeObservableDictionary<string, IStrategy>();
            this._ownOrdersTable = new TradeOrdersTable();
            this._servers = new SmartCOMServerCollection();

            this._workerThread = new EngineWorkerThread(this);

            this._visualBids = null;
            this._visualAsks = null;
            
            ThreadPool.QueueUserWorkItem(delegate
                {
                    Thread.CurrentThread.SetApartmentState(ApartmentState.MTA);
                    Start();
                }
            );

            this._isTestRunning = false;
        }

        /// <summary>
        /// Запуск торговой системы (в нашем случае движка, ядра, поставщика данных)
        /// </summary>
        private void Start()
        {
            try
            {
                this._smartCOMServer = new StServerClass();
                MessageLogger.LogMessage(" Инициализация SmartCOM v " + this._smartCOMServer.GetServerVersionString() + " выполнена успешно.");
            }
            catch (Exception e)
            {
                MessageLogger.LogMessage(" Ошибка инициализации SmartCOM: " + e.Message);
            }

            SetSmartComEventHandlers();
            SetTickerReceiveHandler();

            this._workerThread.StartEngineThread();
        }

        /// <summary>
        /// Установим обработчики получения данных от SmartCOM сервера
        /// </summary>
        private void SetSmartComEventHandlers()
        {
            // обработчики событий соединения
            this._smartCOMServer.Connected    += new _IStClient_ConnectedEventHandler(OnSComConnectedRaised);
            this._smartCOMServer.Disconnected += new _IStClient_DisconnectedEventHandler(OnSComDisconnectRaised);
 
            // обработчики торговых инструментов (Результат GetSymbols)
            this._smartCOMServer.AddSymbol      += new _IStClient_AddSymbolEventHandler(OnAddSymbolRaise);
            this._smartCOMServer.AddTick        += new _IStClient_AddTickEventHandler(OnAddTickRaise);
            this._smartCOMServer.AddBar         += new _IStClient_AddBarEventHandler(OnAddBarRaise);
            this._smartCOMServer.AddTickHistory += new _IStClient_AddTickHistoryEventHandler(OnAddTickHistoryRaise);

            this._smartCOMServer.UpdateBidAsk   += new _IStClient_UpdateBidAskEventHandler(OnAddUpdateBidAskRaise);
            this._smartCOMServer.UpdateQuote    += new _IStClient_UpdateQuoteEventHandler(OnAddUpdateQuoteRaise);

            // работа со счетами 
            this._smartCOMServer.AddPortfolio   += new _IStClient_AddPortfolioEventHandler(OnAddPortfolioRaise);
            this._smartCOMServer.SetPortfolio   += new _IStClient_SetPortfolioEventHandler(OnSetPortfolioRaise);

            //this._smartCOMServer.UpdatePosition += new _IStClient_UpdatePositionEventHandler(OnUpdatePositionRaise);

            //обработчики ответов торговых приказов(ордеров)
            this._smartCOMServer.OrderSucceeded         += new _IStClient_OrderSucceededEventHandler(this.OnOrderPending);
            this._smartCOMServer.OrderFailed            += new _IStClient_OrderFailedEventHandler(this.OnOrderFailed);
            this._smartCOMServer.OrderCancelSucceeded   += new _IStClient_OrderCancelSucceededEventHandler(this.OnOrderCancelSucceeded);
            this._smartCOMServer.OrderCancelFailed      += new _IStClient_OrderCancelFailedEventHandler(this.OnOrderCancelFailed);
            this._smartCOMServer.OrderMoveSucceeded     += new _IStClient_OrderMoveSucceededEventHandler(this.OnOrderMoveSucceed);
            this._smartCOMServer.OrderMoveFailed        += new _IStClient_OrderMoveFailedEventHandler(this.OnOrderMoveFailed);
            this._smartCOMServer.UpdateOrder            += new _IStClient_UpdateOrderEventHandler(this.OnUpdateOrder);
            this._smartCOMServer.AddTrade               += new _IStClient_AddTradeEventHandler(this.OnAddTrade);
        }

        /// <summary>
        /// Установим обработчик получения инструментов (тикеров)
        /// </summary>
        private void SetTickerReceiveHandler()
        {
            // Подписаться на получение инструмента
            this._tickerReceivedRaise = (ticker) =>
                {
                    if (ticker.Symbol == "RTS-6.13_FT")
                    //    (ticker.Code == "GAZR-9.13_FT"))
                    //if (ticker.Code.Contains('-') == true) 
                    //if (ticker.SecExchName == "RTS_FUT")
                    {
                        // подписаться на получение всех сделок
                        //this._smartCOMServer.ListenTicks("RIM3");
                        // подписаться на получение стакана
                        this._smartCOMServer.ListenBidAsks(ticker.Symbol);
                        // подписаться на получение сделок по инструменту
                        this._smartCOMServer.ListenTicks(ticker.Symbol);
                        // подписаться на получение котировок по инструменту
                        this._smartCOMServer.ListenQuotes(ticker.Symbol);
                        
                        // запросить список доступных счетов
                        MessageLogger.LogMessage("Request Accounts");
                        this._smartCOMServer.GetPrortfolioList();

                        
                        ticker.DOM.AssignWithVisualDOM(this._visualBids, this._visualAsks);

                        CreateStrategies();

                        ticker.SubscribeOnDOMUpdate((domArray) =>
                            {
                                Array arrayBids = (Array)domArray.GetValue(0);
                                Array arrayAsks = (Array)domArray.GetValue(1);

                                Action<ObservableCollection<Level>, Array, Action> updateVisualDOM = (visualSnapshot, snapshotArray, updateAction) =>
                                    {
                                        System.Windows.Application.Current.Dispatcher.BeginInvoke(new Action(() =>
                                            {
                                                BaseObjects.SnapshotDOMExtension.SnapshotDOM.CopyArrayToSnapshot(snapshotArray, visualSnapshot);
                                                updateAction();
                                            }));
                                    };

                                updateVisualDOM(this._visualBids, arrayBids, ()=>
                                    {
                                        if (_notifierDOMUpdate != null) _notifierDOMUpdate.NotifyBidsUpdated();
                                    }
                                );
                                updateVisualDOM(this._visualAsks, arrayAsks, () =>
                                    {
                                        if (_notifierDOMUpdate != null) _notifierDOMUpdate.NotifyAsksUpdated();
                                    }
                                );
                                

                            });
                    }
                };
        }

        /// <summary>
        /// Выполняеть соединение с указанным сервером
        /// </summary>
        /// <param name="serverConnection">Параметры соединения с сервером</param>
        /// <param name="login">Логин</param>
        /// <param name="password">Пароль</param>
        public void Connect(SmartCOMServerConnectionParametrs serverConnection, string login, string password)
        {
            this._workerThread.AddEngineTask(() =>
                {
                    //_smartCOMServer.ConfigureServer("pingTimeout=10;logFilePath=log.log");
                    _smartCOMServer.ConfigureClient("logLevel=5");
                    _connectionNotifier.Connection = serverConnection;

                    if (this._smartCOMServer.IsConnected() == false)
                    {
                        this._currentServerConnectionParametrs = serverConnection;
                        this._currentServerConnectionParametrs.Login = login;
                        this._currentServerConnectionParametrs.Password = password;
                        this._smartCOMServer.connect(serverConnection.Ip, serverConnection.Port, login, password);
                    }
                    else
                    {
                        this._isUserDisconnect = true;
                        this.StartStopTest();
                        this._smartCOMServer.disconnect();                        
                    }
                });
        }

        /// <summary>
        /// Проверка работоспособности серверов
        /// </summary>
        public void PingServers()
        {
            this._workerThread.AddEngineTask(() =>
                {
                    Action<SmartCOMServerConnectionParametrs> checkServers = (connectionParametrs) =>
                        {
                            //connectionParametrs.TestOnAvailableSmartComServer(AliveServerMethod.TryPing);
                            connectionParametrs.TestOnAvailableSmartComServer(AliveServerMethod.TryPortOpen);
                        };

                    foreach (var connectionParametrs in this._servers)
                    {
                        Task.Factory.StartNew(() => { checkServers(connectionParametrs); });                            
                    }
                });
        }

        const int numStrategies = 2;

        public void CreateStrategies()
        {
            for (int i = 0; i < numStrategies; i++)
            {
                IStrategy testStrategy = (IStrategy)this.CreateOrGetStrategy(typeof(TestStrategy), "TtS" + i.ToString());
            }
        }


        public void StartStopTest()
        {
            try
            {
                Task[] tasks = new Task[numStrategies];
                for (int i = 0; i < numStrategies; i++)
                {
                    IStrategy testStrategy = (IStrategy)this.GetStrategy("TtS" + i.ToString());
                    if (this._isTestRunning == true)
                    {
                        tasks[i] = testStrategy.Stop();
                    }
                    else
                    {
                        testStrategy.Start();
                    }
                }

                Task.WaitAll(tasks, 20000);
            }
            catch (Exception e)
            {

            }

            this._isTestRunning = !this._isTestRunning;
        }

        #endregion Methods

        #region SmartComEventsHandlers

        protected void OnSComConnectedRaised()
        {
            try
            {
                if (_smartCOMServer.IsConnected() == true)
                {
                    this._connectionNotifier.CurrentConnectionState = Types.ConnectionState.Connected;
                }
                else
                {
                    this._connectionNotifier.CurrentConnectionState = Types.ConnectionState.Disconnected;
                    return;
                }

                // после успешного соединения с сервером, запрашиваем список доступных торговых инструментов
                MessageLogger.LogMessage("Request Symbols");
                this._smartCOMServer.GetSymbols();
            }
            catch (Exception e)
            {
                MessageLogger.LogMessage(e.Message);
            }
        }

        protected void OnSComDisconnectRaised(string reason)
        {
            this._connectionNotifier.CurrentConnectionState = ConnectionState.Disconnected;

            if (this._isUserDisconnect == false)
            {// произведем реконнект в случае, если пользователь не  посылал запрос на разрыв
                this._workerThread.AddEngineTask(() =>
                {
                    if (this._smartCOMServer.IsConnected() == false)
                    {
                        this._smartCOMServer.connect(
                            this._currentServerConnectionParametrs.Ip,
                            this._currentServerConnectionParametrs.Port,
                            this._currentServerConnectionParametrs.Login,
                            this._currentServerConnectionParametrs.Password);
                    }
                });
            }
        }


        private void OnAddSymbolRaise(int row, int nrows, string symbol, string shortName, string longName, 
                                      string type, int decimals, int lotSize, double stepPrice, double step, 
                                      string secExtId, string secExchName, DateTime expiryDate, 
                                      double daysBeforeExpiry, double strike)
        {
            //try
            {
                // добавить инструмент в список
                Ticker tmpTicker = this._instruments.AddOrUpdate(
                    symbol,
                    new Func<string, Ticker>( (name) => 
                    {
                        Ticker ticker = new Ticker(symbol, shortName, longName, step, stepPrice, decimals, 
                                                   secExtId, expiryDate, daysBeforeExpiry);
                        // добавить площадку в список
                        Exchange tmpExchange = this._excangeInstruments.AddOrUpdate(
                            secExchName,
                            new Func<string, Exchange>( (exchangeName) => 
                                {
                                    Exchange tmpExchangeList = new Exchange(secExchName);
                                    return tmpExchangeList;                                
                                }
                            ),

                            new Func<string, Exchange, Exchange>( (exchangeName, prevExchange) => 
                                {
                                    return prevExchange;
                                }
                            )
                        );
                    
                        // привяжем инструмент к площадке
                        tmpExchange.Add(symbol, ticker);

                        return ticker;
                    }),
                
                    new Func<string, Ticker, Ticker>( (name, previousTicker) => 
                    {
                        if (previousTicker == null)
                        {
                            return previousTicker;
                        }
                        previousTicker.Update(step, stepPrice, decimals, expiryDate, daysBeforeExpiry);
                        return previousTicker;
                    })
                );
            
                this._tickerReceivedRaise(tmpTicker);

            }
            //catch (Exception e)
            {
                //MessageLogger.LogMessage(e.Message);
            }
        }

        private void OnAddTickRaise(string symbol, DateTime time, double price, double volume, string dealId, StOrder_Action action)
        {
            Ticker currentTicker = null;
            if (this._instruments.TryGetValue(symbol, out currentTicker) == true)
            {
                Side dealSide = Side.Sell;
                if ((action == StOrder_Action.StOrder_Action_Buy) ||
                    (action == StOrder_Action.StOrder_Action_Buy))
                {

                    dealSide = Side.Buy;
                }
                currentTicker.AddDeal(price, ((int)dealSide) * volume, dealId, time);
            }
            else
            {
                MessageLogger.LogMessage("OnAddUpdateQuoteRaise: Инструмент не найден - " + symbol);
                //throw new Exception("OnAddUpdateQuoteRaise: Инструмент не найден - " + symbol);
            }            
        }
        
        private void OnAddUpdateBidAskRaise(
            string symbol, int row, int nRows, 
            double bidPrice, double bidSize, 
            double askPrice, double askSize)
        {
            Ticker currentTicker = null;
            if (this._instruments.TryGetValue(symbol, out currentTicker) == true)
            {
                try
                {
                    currentTicker.UpdateBidsAsks(row, nRows, bidPrice, bidSize, askPrice, askSize);
                }
                catch (Exception e)
                {
                    MessageLogger.LogMessage(e.Message);
                }
            }
            else
            {
                MessageLogger.LogMessage("OnAddUpdateQuoteRaise: Инструмент не найден - " + symbol);
                //throw new Exception("OnAddUpdateQuoteRaise: Инструмент не найден - " + symbol);
            }            
        }
        
        private void OnAddUpdateQuoteRaise(
            string symbol, DateTime time, double open, double high, double low, double close, double last, double volume, 
            double size, double bid, double ask, double bidSize, double askSize, double openPosition, double goBuy, double goSell, 
            double goBase, double goBaseBacked, double highLimit, double lowLimit, int tradingStatus, double volat, double theoreticPrice)
        {
            try
            {
                // добавить инструмент в список
                Ticker currentTicker = null;
                if (this._instruments.TryGetValue(symbol, out currentTicker) == true)
                {
                    currentTicker.Common.UpdateQuote(highLimit, lowLimit, openPosition, goBuy, goSell, goBase, goBase); 
                }
                else
                {
                    MessageLogger.LogMessage("OnAddUpdateQuoteRaise: Инструмент не найден - " + symbol);
                    //throw new Exception("OnAddUpdateQuoteRaise: Инструмент не найден - " + symbol);
                }
            }
            catch (Exception e)
            {
                MessageLogger.LogMessage(e.Message);
            }
        }


        private void OnAddBarRaise(
            int row, int nrows, string symbol, StBarInterval interval, DateTime datetime, 
            double open, double high, double low, double close, double volume, double open_int)
        {

        }

        private void OnAddTickHistoryRaise(
            int row, int nrows, string symbol, System.DateTime datetime, double price, 
            double volume, string tradeno, StOrder_Action action)
        {
            /*if (nrows == 0)
                Writers.WriteLine("History", "Ticks", "{0} Empty buffer", symbol);
            else
                Writers.WriteLine("History", "Ticks", "{0}/{1} {2} {3} Price: {4} ({5}) {6} {7}", row, nrows, datetime, symbol, price, volume, tradeno, action);
            */
        }


        private void OnAddPortfolioRaise(int row, int nrows, string portfolioName, string portfolioExch, StPortfolioStatus portfolioStatus)
        {
            try
            {
                // добавить площадку в список
                Exchange currentExchange = this._excangeInstruments.GetOrAdd(
                    portfolioExch,
                    new Func<string, Exchange>((exchangeName) =>
                        {
                            Exchange tmpExchange = new Exchange(portfolioExch);
                            return tmpExchange;
                        }
                    )                
                );

                this._accounts.AddOrUpdate(
                    portfolioName, 
                    new Func<string, Account> ((accName) => 
                        {
                            Account tmpAccount = new Account(portfolioName, currentExchange, (PortfolioStatus)portfolioStatus);
                            MessageLogger.LogMessage("Добавлен новый счёт: " + tmpAccount.ToString());                        
                            return tmpAccount;
                        }
                    ), 
                    new Func<string, Account, Account>( (accName, prevAccount) =>
                        {
                            prevAccount.PortfolioStatus = (PortfolioStatus)portfolioStatus;
                            MessageLogger.LogMessage("Обновлена информация по счёту: " + prevAccount.ToString());
                            return prevAccount;
                        }
                    )
                );

                this._smartCOMServer.ListenPortfolio(portfolioName);
            }
            catch (Exception e)
            {
                MessageLogger.LogMessage(e.Message);
            }
        }

        private void OnSetPortfolioRaise(string portfolio, double cash, double leverage, double comission, double saldo)
        {
            try
            {
                Account currentAccount = null;
                if (this._accounts.TryGetValue(portfolio, out currentAccount) == true)
                {
                    currentAccount.UpdatePortfolio(cash, leverage, comission, saldo);
                    MessageLogger.LogMessage("Обновлена информация по счёту: " + currentAccount.ToString());
                }
            }
            catch (Exception e)
            {
                MessageLogger.LogMessage(e.Message);
            }
        }

        private void OnUpdatePositionRaise(string portfolio, string symbol, double avprice, double amount, double planned)
        {
            //Writers.WriteLine("Portfolio", "Position", "{0} {1}:{2} Avg: {3} ({4}) Planed: {5}", DateTime.Now, portfolio, symbol, avprice, amount, planned);
        }

        // Обработчики ответов Заявок

        // Pending ответ
        private void OnOrderPending(int extId, string orderId)
        {
            TradeOrder tmpOrder = this._ownOrdersTable.GetOrder(extId);
            if (tmpOrder == null) return;            

            System.Diagnostics.Debug.WriteLine("orderId=" + orderId + " Pending - " + (DateTime.Now - tmpOrder.Time.Value).Milliseconds.ToString() + " ms");

            Action strategyAction = () =>
            {
                TradeOrder order = tmpOrder;
                string id = orderId;
                
                this._ownOrdersTable.AddOrder(id, order);

                order.AddHistoryCommand(OrderCommand.Pending, orderId);

                this.OrderPending(tmpOrder, orderId);
            };
            tmpOrder.Strategy.SendStrategyAction(strategyAction);
        }

        // Обработка Pending ответа (исполняеться в потоке стратегии)
        private void OrderPending(TradeOrder order, string orderId)
        {
            if (order.OrderId == orderId)
            {
                //событие ордера уже обработано
                return;
            }

            if (order.State == OrderState.Sending)
            {
                order.OrderId = orderId;
                order.State = OrderState.Pending;

                if (order.OnPending != null) order.OnPending(order);
            }
        }

        // Fail  ответ
        private void OnOrderFailed(int extId, string orderId, string reason)
        {
            TradeOrder tmpOrder = this._ownOrdersTable.GetOrder(extId);
            if (tmpOrder == null) return;
            if (tmpOrder.OrderId == orderId) return;

            System.Diagnostics.Debug.WriteLine("orderId" + orderId + "Send Failed - " + (DateTime.Now - tmpOrder.Time.Value).Milliseconds.ToString() + " ms");

            Action strategyAction = () =>
            {
                TradeOrder order = tmpOrder;
                string orderFailReason = reason;

                order.AddHistoryCommand(OrderCommand.Failed, orderFailReason);
                
                this.OrderFailed(tmpOrder, orderFailReason);                
            };
            tmpOrder.Strategy.SendStrategyAction(strategyAction);
        }
        // Обработка Fail ответа (исполняеться в потоке стратегии)
        private void OrderFailed(TradeOrder order, string failReason)
        {            
            if (order.OnFailOpened != null) order.OnFailOpened(order, failReason);

            order.DisposeAndFreeOrder();
        }

        // Cancel  ответ
        private void OnOrderCancelSucceeded(string orderId)
        {
            TradeOrder tmpOrder = this._ownOrdersTable.GetOrder(Convert.ToInt32(orderId));
            if (tmpOrder == null) return;
            if (tmpOrder.OrderId == orderId) return;

            System.Diagnostics.Debug.WriteLine("orderId" + orderId + "Canceled - " + (DateTime.Now - tmpOrder.Time.Value).Milliseconds.ToString() + " ms");

            Action strategyAction = () =>
            {
                TradeOrder order = tmpOrder;
                
                order.AddHistoryCommand(OrderCommand.Canceled, orderId);

                this.OrderCanceled(tmpOrder);
            };
            tmpOrder.Strategy.SendStrategyAction(strategyAction);
        }
        // Обработка Cancel ответа (исполняеться в потоке стратегии)
        private void OrderCanceled(TradeOrder order)
        {            
            if (order.OnCanceled != null) order.OnCanceled(order);
            order.DisposeAndFreeOrder();
        }

        // Cancel Fail ответ
        private void OnOrderCancelFailed(string orderId)
        {            
            TradeOrder tmpOrder = this._ownOrdersTable.GetOrder(orderId);
            if (tmpOrder == null) return;
                //throw new Exception("Пришёл CANCEL FAILED ответ для несуществующего ордера orderId=" + orderId);

            System.Diagnostics.Debug.WriteLine("orderId" + orderId + "Cancel Failed - " + (DateTime.Now - tmpOrder.Time.Value).Milliseconds.ToString() + " ms");

            Action strategyAction = () =>
            {
                TradeOrder order = tmpOrder;

                order.AddHistoryCommand(OrderCommand.CancelFailed, orderId);

                //сбросим последннюю команду
                tmpOrder.ResetLastSendCommand();
                this.OrderCancelFail(tmpOrder);
            };
            tmpOrder.Strategy.SendStrategyAction(strategyAction);

            Array message = new object[] { StrategyDataType.OrderFailCanceled, tmpOrder, orderId };
            tmpOrder.Strategy.SendStrategyData(message);   
        }
        // Обработка Cancel Fail ответа (исполняеться в потоке стратегии)
        private void OrderCancelFail(TradeOrder order)
        {
            if (order.OnFailCanceled != null) order.OnFailCanceled(order);

           // order.DisposeAndFreeOrder();
        }

        // Move  ответ
        private void OnOrderMoveSucceed(string orderId)
        {
            // абсолютно бесполезный ответ, orderId всегда раен "0"
            // ответ от мува обрабатывать будем в событии OnUpdateOrder
            return;
            TradeOrder tmpOrder = this._ownOrdersTable.GetOrder(orderId);
            if (tmpOrder == null) return;
                //throw new Exception("Пришёл MOVE SUCCEED ответ для несуществующего ордера orderId=" + orderId);

            System.Diagnostics.Debug.WriteLine("orderId" + orderId + "Moved - " + (DateTime.Now - tmpOrder.Time.Value).Milliseconds.ToString() + " ms");

            Action strategyAction = () =>
            {
                TradeOrder order = tmpOrder;
                string id = orderId;

                order.AddHistoryCommand(OrderCommand.Moved, orderId);

                this.OrderMoved(tmpOrder, orderId);
            };
            tmpOrder.Strategy.SendStrategyAction(strategyAction);
        }
        // Обработка Move ответа (исполняеться в потоке стратегии)
        private void OrderMoved(TradeOrder order, string orderId)
        {
            if (order.State == OrderState.Moving)
            {
                if (order.LastSendCommand.Item1 != OrderCommand.Move)
                {
                    throw new Exception("Ордер не мувился :" + order.ToString());
                }

                double oldPrice = order.Price;
                double newPrice = (double)order.LastSendCommand.Item2;

                order.ResetLastSendCommand();
                
                order.Price = newPrice;
                //order.OrderId = orderId;
                order.State = OrderState.Open;

                if (order.OnMoved != null) order.OnMoved(order, oldPrice, newPrice);
            }
        }

        // Move Fail  ответ
        private void OnOrderMoveFailed(string orderId)
        {
            // абсолютно бесполезный ответ, orderId всегда раен "0"
            // ответ от мува обрабатывать будем в событии OnUpdateOrder
            return;
            TradeOrder tmpOrder = this._ownOrdersTable.GetOrder(orderId);
            if (tmpOrder == null) return;
                //throw new Exception("Пришёл MOVE FAILED ответ для несуществующего ордера orderId=" + orderId);

            Action strategyAction = () =>
            {
                TradeOrder order = tmpOrder;
                string id = orderId;

                this.OrderFailMoved(tmpOrder, orderId);
            };
            tmpOrder.Strategy.SendStrategyAction(strategyAction);

        }
        // Обработка Move Fail ответа (исполняеться в потоке стратегии)
        private void OrderFailMoved(TradeOrder order, string orderId)
        {
            if (order.State == OrderState.Moving)
            {
                if (order.LastSendCommand.Item1 != OrderCommand.Move)
                {
                    throw new Exception("Ордер не мувился :" + order.ToString());
                }

                double oldPrice = order.Price;
                double newPrice = (double)order.LastSendCommand.Item2;
                order.ResetLastSendCommand();

                order.State = OrderState.Open;

                if (order.OnFailMoved != null) order.OnFailMoved(order, oldPrice, newPrice);
            }
        }

        private void OnUpdateOrder(string portfolio, string symbol, StOrder_State state, StOrder_Action action,
            StOrder_Type type, StOrder_Validity validity, double price, double amount, double stop,
            double notFilledAmount, System.DateTime datetime, string orderId, string exchangeOrderNo, int status_mask, int extId)
        {
            string updateString = "Update " + state.ToString() + " extId=" + extId.ToString() + " \t Id comment " + orderId + " \t exchangeOrderNo  " + exchangeOrderNo +  " \t price= " + price + " \t amount=" + amount + " \t rest=" + notFilledAmount.ToString();

            TradeOrder tmpOrder = null;
            int signSideMul = 1;
            if ((action == StOrder_Action.StOrder_Action_Sell)
              || (action == StOrder_Action.StOrder_Action_Short)) signSideMul = -1;

            if (state == StOrder_State.StOrder_State_Pending)
            {
                tmpOrder = this._ownOrdersTable.GetOrder(extId);
                if (tmpOrder != null)
                {
                    System.Diagnostics.Debug.WriteLine("orderId" + orderId + "Update Pending - " + (DateTime.Now - tmpOrder.Time.Value).Milliseconds.ToString() + " ms");
                    Action strategyActionPending = () =>
                    {
                        TradeOrder order = tmpOrder;
                        string id = orderId;

                        order.AddHistoryCommand(OrderCommand.Pending, "Update extId=" + extId.ToString() + " exchangeOrderNo=" + exchangeOrderNo + " State=" + state.ToString() + " ; " + order.ToString());

                        this.OrderPending(order, id);
                    };
                    tmpOrder.Strategy.SendStrategyAction(strategyActionPending);
                }
            }
            else
            if (state == StOrder_State.StOrder_State_Open)
            {
                tmpOrder = this._ownOrdersTable.GetOrder(extId);
                //if ((tmpOrder != null) && (tmpOrder.ExchangeOrderId == "0"))
                if (tmpOrder != null)
                {
                    System.Diagnostics.Debug.WriteLine("orderId" + orderId + "Update Open - " + (DateTime.Now - tmpOrder.Time.Value).Milliseconds.ToString() + " ms");
                    
                    Action strategyActionOpen = () =>
                    {
                        TradeOrder order = tmpOrder;
                        string id = orderId;

                        this._ownOrdersTable.AddOrder(id, order);

                        string tmpExchangeOrderNo = exchangeOrderNo;
                        double tmpOrderUpdatePrice = price;
                        order.AddHistoryCommand(OrderCommand.Opened, "Update orderId=" + orderId + " ;  exchangeOrderNo=" + exchangeOrderNo + " State=" + state.ToString() + " ; " + order.ToString());

                        this.OrderOpened(order, id, tmpExchangeOrderNo, tmpOrderUpdatePrice);
                    };
                    tmpOrder.Strategy.SendStrategyAction(strategyActionOpen);
                }
            }
            else
            if((state == StOrder_State.StOrder_State_Partial)
              ||(state == StOrder_State.StOrder_State_Filled))
            {
                tmpOrder = this._ownOrdersTable.GetOrder(extId);
                if (tmpOrder != null)
                {   
                    Action strategyActionFill = () => 
                    {
                        TradeOrder order = tmpOrder;

                        double filledAmount = order.Amount - signSideMul * notFilledAmount;
                        double dealAmount = order.AmountRest - signSideMul * notFilledAmount;
                        order.AddHistoryCommand(OrderCommand.OrderFill, "Update Fill orderId"  + orderId + " ;  exchangeOrderNo=" + exchangeOrderNo + " State=" + state.ToString()
                            + " price=" + price + " amount=" + amount + " rest=" + (notFilledAmount * signSideMul).ToString() + " ; " + order.ToString());

                        order.AmountFilled += dealAmount;


                        Deal ownNewDeal = new Deal(order.Account, order.Ticker, price, dealAmount, exchangeOrderNo, datetime);
                        //this._ownDeals.Add(ownNewDeal);

                        if (order.OnPartialFilled != null) order.OnPartialFilled(order, dealAmount);

                        if (order.AmountRest == 0.0)
                        {
                            if (order.OnFilled != null) order.OnFilled(order);
                            order.DisposeAndFreeOrder();
                        }
                    };
                    tmpOrder.Strategy.SendStrategyAction(strategyActionFill); 
                }
            }
            else
            if (state == StOrder_State.StOrder_State_Cancel)
            {
                tmpOrder = this._ownOrdersTable.GetOrder(extId);
                if ((tmpOrder != null) && (tmpOrder.OrderId == orderId))
                {
                    System.Diagnostics.Debug.WriteLine("orderId" + orderId + "Update Cancel - " + (DateTime.Now - tmpOrder.Time.Value).Milliseconds.ToString() + " ms");

                    Action strategyActionCancel = () =>
                    {
                        TradeOrder order = tmpOrder;

                        order.AddHistoryCommand(OrderCommand.Canceled, "Update Canceled orderId=" + orderId + " ;  exchangeOrderNo=" + exchangeOrderNo + " State=" + state.ToString()
                            + " price=" + price + " amount=" + amount + " rest=" + (notFilledAmount * signSideMul).ToString() + " ; " + order.ToString());

                        this.OrderCanceled(order);
                    };
                    tmpOrder.Strategy.SendStrategyAction(strategyActionCancel);
                }
            }
            else
            if ((state == StOrder_State.StOrder_State_Expired)
              ||(state == StOrder_State.StOrder_State_ContragentCancel)
              ||(state == StOrder_State.StOrder_State_ContragentReject)
              ||(state == StOrder_State.StOrder_State_SystemCancel)
              ||(state == StOrder_State.StOrder_State_SystemReject))
            {
                tmpOrder = this._ownOrdersTable.GetOrder(extId);
                if ((tmpOrder != null) && (tmpOrder.OrderId == orderId))
                {
                    System.Diagnostics.Debug.WriteLine("ERROR orderId" + orderId + " " + state.ToString() + " - " + (tmpOrder.Time.Value - DateTime.Now).Milliseconds.ToString() + " ms");
                    System.Diagnostics.Debug.WriteLine("ERROR Update orderId=" + orderId + " ; exchangeOrderNo  " + exchangeOrderNo + " ; " + state.ToString() + " ; " + tmpOrder.State.ToString());
                    Action strategyActionError = () =>
                    {
                        TradeOrder order = tmpOrder;

                        order.AddHistoryCommand(OrderCommand.Error, "ERROR Update orderId=" + orderId + " ;  exchangeOrderNo=" + exchangeOrderNo + " State=" + state.ToString()
                            + " price=" + price + " amount=" + amount + " rest=" + (notFilledAmount * signSideMul).ToString() + " ; " + order.ToString());

                        //if ((order.State == OrderState.Pending) || (order.State == OrderState.Sending))
                        {
                            order.DisposeAndFreeOrder();
                        }
                        
                    };
                    tmpOrder.Strategy.SendStrategyAction(strategyActionError);
                }
            }

            if (notFilledAmount == 0.0) // ордер снят или исполнен
            {
                tmpOrder = this._ownOrdersTable.GetOrder(extId);
                if (tmpOrder != null) 
                {
                    if (tmpOrder.State == OrderState.Canceling)
                    {
                        System.Diagnostics.Debug.WriteLine("ERROR Update orderId=" + orderId + " ; exchangeOrderNo  " + exchangeOrderNo + " ; " + state.ToString() + " ; " + tmpOrder.State.ToString());
                        Action strategyActionError = () =>
                        {
                            TradeOrder order = tmpOrder;

                            order.AddHistoryCommand(OrderCommand.OrderFree, "Update DisposeAndFree orderId=" + orderId + " ; exchangeOrderNo  " + exchangeOrderNo + " ; " + state.ToString() + " ; " + order.State.ToString());

                            order.DisposeAndFreeOrder();
                        };
                        tmpOrder.Strategy.SendStrategyAction(strategyActionError);
                    }
                }                
            }

            if (tmpOrder == null)
            {
                updateString = "NULL ORDER " + updateString;
            }
            System.Diagnostics.Debug.WriteLine(updateString);
            
        }

        // Обработка Open ответа (исполняеться в потоке стратегии)
        private void OrderOpened(TradeOrder order, string orderId, string exchangeOrderId= null, double orderUpdatePrice = 0.0)
        {
            if (exchangeOrderId != null)
            {
                if (order.ExchangeOrderId == "0") order.ExchangeOrderId = exchangeOrderId;
            }

            if ((order.State == OrderState.Sending) || (order.State == OrderState.Pending))
            {                
                order.State = OrderState.Open;
                order.OrderId = orderId;
                if (order.OnOpened != null) order.OnOpened(order);
            }
            else
            if (order.State == OrderState.Moving)
            {
                if (exchangeOrderId != null)
                {
                    long newExchangeOrderId = Int64.Parse(exchangeOrderId);
                    long currentExchangeOrderId = Int64.Parse(order.ExchangeOrderId);

                    if ((order.OrderId == orderId) && (newExchangeOrderId > currentExchangeOrderId))
                    {
                        order.ExchangeOrderId = exchangeOrderId;

                        if (order.LastSendCommand.Item1 != OrderCommand.Move)
                        {
                            throw new Exception("Ордер не мувился :" + order.ToString());
                        }


                        double oldPrice = order.Price;
                        double newPrice = (double)order.LastSendCommand.Item2;

                        if (newPrice != orderUpdatePrice)
                        {
                            return;
                        }

                        order.ResetLastSendCommand();

                        order.Price = newPrice;
                        //order.OrderId = orderId;
                        order.State = OrderState.Open;

                        if (order.OnMoved != null) order.OnMoved(order, oldPrice, newPrice);

                    }
                    else
                    {

                    }
                }
                else                
                {

                }
            }
            else
            {

            }

        }

        /// <summary>
        /// Событие о сделке (исполнении ордера с биржевым номером exchangeId = orderNo;
        /// </summary>
        private void OnAddTrade(string account, string symbol, string orderNo, double price, double amount, DateTime datetime, string tradeno)
        {
            TradeOrder tmpOrder = null;
            var orders = this._ownOrdersTable.GetFilteredOrders( (preFilteredOrder) => 
                {
                    return preFilteredOrder.ExchangeOrderId == orderNo;
                });
            if (orders.Count > 0)
            {
                tmpOrder = orders[0];
            }

            Action strategyActionFill = () =>
            {
                TradeOrder order = tmpOrder;
                double dealAmount = amount;
                double dealPrice = price;
                string dealTradeNumber = tradeno;
                string dealOrderNo = orderNo;
                DateTime dealTime = datetime;
                order.AddHistoryCommand(OrderCommand.OrderFill, "AddTrade Fill  ; exchangeOrderNo  " + dealOrderNo + " ; " + order.State.ToString() + " price; amount/notFilledAmount" + dealPrice + "/" + dealAmount + "/" + order.AmountRest);

                order.AmountFilled += dealAmount;

                Deal ownNewDeal = new Deal(tmpOrder.Account, tmpOrder.Ticker, dealPrice, dealAmount, dealTradeNumber, dealTime);
                //this._ownDeals.Add(ownNewDeal);

                if (order.OnPartialFilled != null) order.OnPartialFilled(order, dealAmount);

                if (order.AmountRest == 0.0)
                {
                    if (order.OnFilled != null) order.OnFilled(order);
                    order.DisposeAndFreeOrder();
                }
            };

            System.Diagnostics.Debug.WriteLine("AddTrade Fill  exchangeOrderNo  " + orderNo + " ; " + "price " + amount + " ; amount" + amount);

            return;
            if (tmpOrder != null)
            {                
                tmpOrder.Strategy.SendStrategyAction(strategyActionFill);
            }
            else
            {                
                int startHour = ExtIdGenerator.StartTime.Hour + 0; // здесь необходимо учесть часовые пояса
                int startMinute = ExtIdGenerator.StartTime.Minute;
                int startSecond = ExtIdGenerator.StartTime.Second;

                if ((datetime.Hour >= startHour) &&
                    (datetime.Minute >= startMinute) &&
                    (datetime.Second >= startSecond))
                {
                    // необходимо дожнаться получения идентификатора биржи по заявке, чтоб испполнение привязать к нему
                    // попытаемся поискать сделку еще раз
                    this._workerThread.AddEngineTask( new Action( ()=>
                        {
                            Action waitOrderExhangeId = null;
                            double dealAmount = amount;
                            double dealPrice = price;
                            string dealTradeNumber = tradeno;
                            string dealOrderNo = orderNo;
                            DateTime dealTime = datetime;

                            waitOrderExhangeId = new Action(() =>
                            {
                                TradeOrder order = null;
                                var ordersInWaitAction = this._ownOrdersTable.GetFilteredOrders((preFilteredOrder) =>
                                {
                                    return preFilteredOrder.ExchangeOrderId == orderNo;
                                });
                                if (orders.Count > 0)
                                {
                                    order = orders[0];
                                }

                                if (order == null)
                                {
                                    this._workerThread.AddEngineTask(waitOrderExhangeId);
                                }
                                else
                                {
                                    System.Diagnostics.Debug.WriteLine("AddTrade Fill  exchangeOrderNo  " + orderNo + " ; " + tmpOrder.State.ToString() + " ; rest" + tmpOrder.AmountRest);
                                    tmpOrder.Strategy.SendStrategyAction(strategyActionFill);
                                }
                            });

                            waitOrderExhangeId();
                        }));
                }
                // по данным сделки не определить какие стратегии имеют позицию.
                // Поэтому пропустим сделки. Но в реале эти сделки имеет смысл получить, для того чтобы закрыть открытые ранее позиции
                return;

                // Необходимо для инициализации стратегий после перезапуска торговой ссиетмы,чтобы прогрузить все сделки
                // и выщитать текущую позицию по отдельным стратегиям
                Account tmpAccount = null;
                Ticker tmpTicker = null;
                Deal ownNewDeal = null;
                if ((this._accounts.TryGetValue(account, out tmpAccount) == true) &&
                    (this._instruments.TryGetValue(symbol, out tmpTicker) == true))
                {
                    ownNewDeal = new Deal(tmpAccount, tmpTicker, price, amount, tradeno, datetime);
                }

                //Тестовое заполнение сделок
                BaseStrategy str = (BaseStrategy)this.Strategies["TtS0"];

                int extId = ExtIdGenerator.GetExtId();
                TradeOrder ord = TradeOrder.GetInitializedOrder(str, tmpAccount, tmpTicker, price, amount, extId);

                Action addOrderAction = () =>
                    {
                        TradeOrder orderToAdd = ord;
                        str.Orders.AddOrder(tradeno, orderToAdd);
                    };

                var message = new object[] { StrategyDataType.SGH, addOrderAction };
                this.Strategies["TtS0"].SendStrategyData(message);

                message = new object[] { StrategyDataType.OrderPartialFilled, ord, ownNewDeal };
                str.SendStrategyData(message);

            }
        }


        #endregion SmartComEventsHandlers

        #region Properties

        public ServerConnectionStateNotifier<SmartCOMServerConnectionParametrs> ConnectionNotifier
        {
            get { return this._connectionNotifier; }
        }

        public bool IsConnected { get { return this._smartCOMServer.IsConnected(); } }

        public TradeOrdersTable OwnOrders { get { return this._ownOrdersTable; } }

        public SmartCOMServerCollection Servers { get { return this._servers; } }

        public SafeObservableDictionary<string, IStrategy> Strategies { get { return this._strategies; } }
        
        #endregion Properties

        #region InterfaceImplementation

        /// <summary>
        /// Получить биржу по имени
        /// </summary>
        /// <param name="exchangeName">Имя биржи</param>
        /// <returns>Экземпляр биржи</returns>
        public Exchange GetExchange(string exchangeName)
        {
            Exchange tmpExchange = null;
            this._excangeInstruments.TryGetValue(exchangeName, out tmpExchange);
            return tmpExchange;
        }

        /// <summary>
        /// Получить торговый инструмент
        /// </summary>
        /// <param name="ticker">Наименование инструмента</param>
        /// <returns>"Экемпляр инструмента</returns>
        public Ticker GetTicker(string tickerName)
        {
            Ticker tmpTicker = null;
            this._instruments.TryGetValue(tickerName, out tmpTicker);
            return tmpTicker;
        }

        /// <summary>
        /// Получить торговый счет
        /// </summary>
        /// <param name="exchange">Наименование счёта</param>
        /// <returns>Экземпляр счёта</returns>
        public List<Account> GetAccounts(Exchange exchange)
        {
            return new List<Account>(
                        from acc in this._accounts
                        where acc.Value.PortfolioExchange.Name == exchange.Name
                        select acc.Value);
        }

        /// <summary>
        /// Получить экземпляр стратегии
        /// </summary>
        /// <param name="name">Имя стратегии</param>
        /// <returns>Экземпляр стратегии</returns>
        public IStrategy GetStrategy(string name)
        {
            IStrategy tmpStrategy = null;
            this._strategies.TryGetValue(name, out tmpStrategy);
            return tmpStrategy;
        }

        /// <summary>
        /// Создать стратегию
        /// </summary>
        /// <param name="strategyType">Тип Класса создаваемой стратегии</param>
        /// <param name="name">Имя стратегии</param>
        /// <returns>"Экземпляр стратегии</returns>
        public IStrategy CreateOrGetStrategy(Type strategyType, string name)
        {
            IStrategy newStrategy = null;
            if (this._strategies.TryGetValue(name, out newStrategy) == true) return newStrategy;
            newStrategy = (IStrategy)Activator.CreateInstance(strategyType, new object[] { this, name });
            if (this._strategies.TryAdd(name, (IStrategy)newStrategy) == true)
            {
                return (IStrategy)newStrategy;
            }
            return null;
        }

        /// <summary>
        /// Отправить Заявку с параметрами из экземпляра order
        /// </summary>
        /// <param name="order">Экземпляр торгового ордера</param>
        public void SendOrder(TradeOrder order)
        {
            this._ownOrdersTable.AddOrder(order.ExtId, order);

            StOrder_Action orderAPISide = StOrder_Action.StOrder_Action_Buy;
            if (order.Side == Side.Sell) orderAPISide = StOrder_Action.StOrder_Action_Sell;

            StOrder_Type orderAPIType = StOrder_Type.StOrder_Type_Limit;
            StOrder_Validity orderAPIValidity = StOrder_Validity.StOrder_Validity_Day;

            if (order.Type == OrderType.LimitGTD) orderAPIValidity = StOrder_Validity.StOrder_Validity_Gtc;
            if (order.Type == OrderType.Market) orderAPIType = StOrder_Type.StOrder_Type_Market;
            
            //if (stop != 0.0)
            {
                if (order.Type == OrderType.Stop) orderAPIType = StOrder_Type.StOrder_Type_Stop;
                if (order.Type == OrderType.StopLimit) orderAPIType = StOrder_Type.StOrder_Type_StopLimit;
                if (order.Type == OrderType.StopLimitGTD)
                {
                    orderAPIType = StOrder_Type.StOrder_Type_StopLimit;
                    orderAPIValidity = StOrder_Validity.StOrder_Validity_Gtc;
                }
            }
            
            DateTime sendTime = DateTime.Now;
            this._smartCOMServer.PlaceOrder(order.Account.Name, order.Ticker.Symbol, orderAPISide, orderAPIType, orderAPIValidity,
                order.Price, Math.Abs(order.Amount), order.StopValue, order.ExtId);
            order.Time = sendTime;
        }
        
        /// <summary>
        /// Перемещение заявки
        /// </summary>
        /// <param name="order">Ордер для перемещения</param>
        /// <param name="newPrice">Новое значение цены</param>
        public void MoveOrder(TradeOrder order, double newPrice)
        {
            DateTime sendTime = DateTime.Now;

            System.Diagnostics.Debug.WriteLine("MOVE extId=" + order.ExtId.ToString() + " ;  exchangeOrderNo  " + order.ExchangeOrderId + " Id comment " + order.OrderId + " ; " + order.State.ToString() + "; price Newprice  amount/notFilledAmount" + order.Price + " / " + newPrice + "/" + order.Amount + "/" + order.AmountRest);
            
            this._smartCOMServer.MoveOrder(order.Account.Name, order.OrderId, newPrice);
            order.Time = sendTime;
        }

        /// <summary>
        /// Снятие заявки order
        /// </summary>
        /// <param name="order">отменяемый(снимаемый) торговый ордер</param>
        public void CancelOrder(TradeOrder order)
        {
            System.Diagnostics.Debug.WriteLine("CANCEL extId=" + order.ExtId.ToString() + " ;  exchangeOrderNo  " + order.ExchangeOrderId + " Id comment " + order.OrderId + " ; " + order.State.ToString() + "; price amount/notFilledAmount" + order.Price + " / " + "/" + order.Amount + "/" + order.AmountRest);
            DateTime sendTime = DateTime.Now;
            this._smartCOMServer.CancelOrder(order.Account.Name, order.Ticker.Symbol ,order.OrderId);
            order.Time = sendTime;
        }

        #endregion InterfaceImplementation
    }
}
